"""Mesh topology representation."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import pytools
import numpy
import numpy.linalg as la

# make sure AffineMap monkeypatch happens
import hedge.tools




class TAG_NONE(object): 
    """A boundary or volume tag representing an empty boundary or volume."""
    pass
class TAG_ALL(object): 
    """A boundary or volume tag representing the entire boundary or volume."""
    pass
class TAG_NO_BOUNDARY(object): 
    """A boundary tag indicating that this edge should not fall under TAG_ALL."""
    pass
class TAG_RANK_BOUNDARY(object): 
    """A boundary tag indicating the boundary with a neighboring rank."""
    def __init__(self, rank):
        self.rank = rank

    def __eq__(self, other):
        return isinstance(other, TAG_RANK_BOUNDARY) and self.rank == other.rank

    def __hash__(self):
        return 0xaffe ^ hash(self.rank)






def make_element(class_, id, vertex_indices, all_vertices):
    vertices = [all_vertices[v] for v in vertex_indices]
    map = class_.get_map_unit_to_global(vertices)
    new_vertex_indices = \
            class_._reorder_vertices(vertex_indices, vertices, map)
    if new_vertex_indices:
        vertex_indices = new_vertex_indices
        vertices = [all_vertices[v] for v in vertex_indices]
        map = class_.get_map_unit_to_global(vertices)

    face_normals, face_jacobians = \
            class_.face_normals_and_jacobians(vertices, map)

    return class_(id, vertex_indices, map, map.inverted(), 
            face_normals, face_jacobians)




class Element(pytools.Record):
    __slots__ = ["id", "vertex_indices", "map", "inverse_map", "face_normals",
            "face_jacobians"]

    def __init__(self, id, vertex_indices, map, inverse_map, face_normals,
            face_jacobians):
        pytools.Record.__init__(self, locals())

    @staticmethod
    def _reorder_vertices(vertex_indices, vertices):
        return vertex_indices

    def bounding_box(self, vertices):
        my_verts = numpy.array([vertices[vi] for vi in self.vertex_indices])
        return numpy.min(my_verts, axis=0), numpy.max(my_verts, axis=0)

    def centroid(self, vertices):
        my_verts = numpy.array([vertices[vi] for vi in self.vertex_indices])
        return numpy.average(my_verts, axis=0)






class SimplicialElement(Element):
    __slots__ = ["id", "vertex_indices", "map", "inverse_map", "face_normals",
            "face_jacobians"]

    @property
    def faces(self):
        return self.face_vertices(self.vertex_indices)

    @classmethod
    def get_map_unit_to_global(cls, vertices):
        """Return an affine map that maps the unit coordinates of the reference
        element to a global element at a location given by its `vertices'.
        """
        from hedge._internal import get_simplex_map_unit_to_global
        return get_simplex_map_unit_to_global(cls.dimensions, vertices)

    def contains_point(self, x):
        unit_coords = self.inverse_map(x)
        for xi in unit_coords:
            if xi < -1:
                return False
        return sum(unit_coords) <= -(self.dimensions-2)




class Interval(SimplicialElement):
    dimensions = 1
    @staticmethod
    def face_vertices(vertices):
        return [(vertices[0],), (vertices[1],) ]

    @classmethod
    def _reorder_vertices(cls, vertex_indices, vertices, map):
        vi = vertex_indices
        if vertices[0][0] > vertices[1][0]: # make sure we're ordered left-right
            return (vi[1], vi[0])
        else:
            return None

    @staticmethod
    def face_normals_and_jacobians(vertices, affine_map):
        """Compute the normals and face jacobians of the unit element
        transformed according to `affine_map'.

        Returns a pair of lists [normals], [jacobians].
        """
        if affine_map.jacobian < 0:
            return [
                    numpy.array([1], dtype=float), 
                    numpy.array([-1], dtype=float)
                    ], [1, 1]
        else:
            return [
                    numpy.array([-1], dtype=float), 
                    numpy.array([1], dtype=float)
                    ], [1, 1]




class Triangle(SimplicialElement):
    dimensions = 2

    __slots__ = ["id", "vertex_indices", "map", "inverse_map", "face_normals",
            "face_jacobians"]

    @staticmethod
    def face_vertices(vertices):
        return [(vertices[0], vertices[1]), 
                (vertices[1], vertices[2]), 
                (vertices[0], vertices[2])
                ]

    @classmethod
    def _reorder_vertices(cls, vertex_indices, vertices, map):
        vi = vertex_indices
        if map.jacobian > 0:
            return (vi[0], vi[2], vi[1])
        else:
            return None

    @staticmethod
    def face_normals_and_jacobians(vertices, affine_map):
        """Compute the normals and face jacobians of the unit element
        transformed according to `affine_map'.

        Returns a pair of lists [normals], [jacobians].
        """
        from hedge.tools import sign

        m = affine_map.matrix
        orient = sign(affine_map.jacobian)
        face1 = m[:,1] - m[:,0]
        raw_normals = [
                orient*numpy.array([m[1,0], -m[0,0]]),
                orient*numpy.array([face1[1], -face1[0]]),
                orient*numpy.array([-m[1,1], m[0,1]]),
                ]

        face_lengths = [numpy.linalg.norm(fn) for fn in raw_normals]
        return [n/fl for n, fl in zip(raw_normals, face_lengths)], \
                face_lengths




class Tetrahedron(SimplicialElement):
    dimensions = 3

    __slots__ = ["id", "vertex_indices", "map", "inverse_map", "face_normals",
            "face_jacobians"]

    @staticmethod
    def face_vertices(vertices):
        return [(vertices[0],vertices[1],vertices[2]), 
                (vertices[0],vertices[1],vertices[3]),
                (vertices[0],vertices[2],vertices[3]),
                (vertices[1],vertices[2],vertices[3]),
                ]

    @classmethod
    def _reorder_vertices(cls, vertex_indices, vertices, map):
        vi = vertex_indices
        if map.jacobian > 0:
            return (vi[0], vi[1], vi[3], vi[2])
        else:
            return None

    @classmethod
    def face_normals_and_jacobians(cls, vertices, affine_map):
        """Compute the normals and face jacobians of the unit element
        transformed according to `affine_map'.

        Returns a pair of lists [normals], [jacobians].
        """
        from hedge.tools import normalize, sign

        face_orientations = [-1,1,-1,1]
        element_orientation = sign(affine_map.jacobian)

        def fj_and_normal(fo, pts):
            normal = numpy.cross(pts[1]-pts[0], pts[2]-pts[0])
            n_length = la.norm(normal)

            # ||n_length|| is the area of the parallelogram spanned by the two
            # vectors above. Half of that is the area of the triangle we're interested
            # in. Next, the area of the unit triangle is two, so divide by two again.
            return element_orientation*fo*normal/n_length, n_length/4

        m = affine_map.matrix

        # realize that zip(*something) is unzip(something)
        return zip(*[fj_and_normal(fo, pts) for fo, pts in
            zip(face_orientations, cls.face_vertices(vertices))])




class Mesh(pytools.Record):
    """Information about the geometry and connectivity of a finite
    element mesh. (Note: no information about the discretization
    is stored here.)

    @ivar points: list of Pylinear vectors of node coordinates
    @ivar elements: list of Element instances
    @ivar interfaces: a list of pairs::

          ((element instance 1, face index 1), (element instance 2, face index 2))

      enumerating elements bordering one another.  The relation "element 1 touches 
      element 2" is always reflexive, but this list will only contain one entry
      per element pair.
    @ivar tag_to_boundary: a mapping of the form::
          boundary_tag -> [(element instance, face index)])

      The boundary tag TAG_NONE always refers to an empty boundary.
      The boundary tag TAG_ALL always refers to the entire boundary.
    @ivar tag_to_elements: a mapping of the form
      element_tag -> [element instances]

      The boundary tag TAG_NONE always refers to an empty domain.
      The boundary tag TAG_ALL always refers to the entire domain.
    @ivar periodicity: A list of tuples (minus_tag, plus_tag) or None
      indicating the tags of the boundaries to be matched together
      as periodic. There is one tuple per axis, so that for example
      a 3D mesh has three tuples.
    @ivar periodic_opposite_faces: a mapping of the form::
          (face_vertex_indices) -> 
            (opposite_face_vertex_indices), axis

      This maps a face to its periodicity-induced opposite.

    @ivar periodic_opposite_vertices: a mapping of the form::
          vertex_index -> [(opposite_vertex_index, axis), ...]

      This maps one vertex to a list of its periodicity-induced 
      opposites.
    """

    def both_interfaces(self):
        for face1, face2 in self.interfaces:
            yield face1, face2
            yield face2, face1

    def bounding_box(self):
        try:
            return self._bounding_box
        except AttributeError:
            self._bounding_box = (
                    numpy.min(self.points, axis=0),
                    numpy.max(self.points, axis=0),
                    )
            return self._bounding_box

    def element_adjacency_graph(self):
        """Return a dictionary mapping each element id to a
        list of adjacent element ids.
        """
        adjacency = {}
        for (e1, f1), (e2, f2) in self.interfaces:
            adjacency.setdefault(e1.id, []).append(e2.id)
            adjacency.setdefault(e2.id, []).append(e1.id)
        return adjacency





def _build_mesh_data_dict(points, elements, boundary_tagger, periodicity, is_rankbdry_face):
    # create face_map, which is a mapping of
    # (vertices on a face) -> 
    #  [(element, face_idx) for elements bordering that face]
    face_map = {}
    for el in elements:
        for fid, face_vertices in enumerate(el.faces):
            face_map.setdefault(frozenset(face_vertices), []).append((el, fid))

    # build non-periodic connectivity structures
    interfaces = []
    tag_to_boundary = {TAG_NONE: [], TAG_ALL: []}
    for face_vertices, els_faces in face_map.iteritems():
        if len(els_faces) == 2:
            interfaces.append(els_faces)
        elif len(els_faces) == 1:
            el, face = els_faces[0]
            tags = boundary_tagger(face_vertices, el, face)

            if isinstance(tags, str):
                from warnings import warn
                warn("Received string as tag list")

            for btag in tags:
                tag_to_boundary.setdefault(btag, []) \
                        .append(els_faces[0])
            if TAG_NO_BOUNDARY not in tags:
                # this is used to mark rank interfaces as not being part of the
                # boundary
                tag_to_boundary[TAG_ALL].append(els_faces[0])
        else:
            raise RuntimeError, "face can at most border two elements"

    # add periodicity-induced connectivity
    from pytools import flatten, reverse_dictionary

    periodic_opposite_faces = {}
    periodic_opposite_vertices = {}

    for axis, axis_periodicity in enumerate(periodicity):
        if axis_periodicity is not None:
            # find faces on +-axis boundaries
            minus_tag, plus_tag = axis_periodicity
            minus_faces = tag_to_boundary.get(minus_tag, [])
            plus_faces = tag_to_boundary.get(plus_tag, [])

            # find vertex indices and points on these faces
            minus_vertex_indices = list(set(flatten(el.faces[face] 
                for el, face in minus_faces)))
            plus_vertex_indices = list(set(flatten(el.faces[face] 
                for el, face in plus_faces)))

            minus_z_points = [points[pi] for pi in minus_vertex_indices]
            plus_z_points = [points[pi] for pi in plus_vertex_indices]

            # find a mapping from -axis to +axis vertices
            from hedge.tools import find_matching_vertices_along_axis

            minus_to_plus, not_found = find_matching_vertices_along_axis(
                    axis, minus_z_points, plus_z_points,
                    minus_vertex_indices, plus_vertex_indices)
            plus_to_minus = reverse_dictionary(minus_to_plus)

            for a, b in minus_to_plus.iteritems():
                periodic_opposite_vertices.setdefault(a, []).append((b, axis))
                periodic_opposite_vertices.setdefault(b, []).append((a, axis))

            # establish face connectivity
            for minus_face in minus_faces:
                minus_el, minus_fi = minus_face
                minus_fvi = minus_el.faces[minus_fi]

                try:
                    mapped_plus_fvi = tuple(minus_to_plus[i] for i in minus_fvi)
                    plus_faces = face_map[frozenset(mapped_plus_fvi)]
                    assert len(plus_faces) == 1
                except KeyError:
                    # is our periodic counterpart is in a different mesh clump?
                    if is_rankbdry_face(minus_face):
                        # if so, cool. parallel handler will take care of it.
                        continue
                    else:
                        # if not, bad.
                        raise

                plus_face = plus_faces[0]
                interfaces.append([minus_face, plus_face])

                plus_el, plus_fi = plus_face
                plus_fvi = plus_el.faces[plus_fi]

                mapped_minus_fvi = tuple(plus_to_minus[i] for i in plus_fvi)

                # periodic_opposite_faces maps face vertex tuples from
                # one end of the periodic domain to the other, while
                # correspondence between each entry 

                periodic_opposite_faces[minus_fvi] = mapped_plus_fvi, axis
                periodic_opposite_faces[plus_fvi] = mapped_minus_fvi, axis

                tag_to_boundary[TAG_ALL].remove(plus_face)
                tag_to_boundary[TAG_ALL].remove(minus_face)
    return {
            "interfaces": interfaces,
            "tag_to_boundary": tag_to_boundary,
            "periodicity": periodicity,
            "periodic_opposite_faces": periodic_opposite_faces,
            "periodic_opposite_vertices": periodic_opposite_vertices,
            }




def make_conformal_mesh(points, elements, 
        boundary_tagger=(lambda fvi, el, fn: []), 
        element_tagger=(lambda el: []),
        periodicity=None,
        _is_rankbdry_face=(lambda (el, face): False),
        ):
    """Construct a simplical mesh.

    Face indices follow the convention for the respective element,
    such as Triangle or Tetrahedron, in this module.

    @param points: an iterable of vertex coordinates, given as vectors.
    @param elements: an iterable of tuples of indices into points,
      giving element endpoints.
    @param boundary_tagger: a function that takes the arguments
      C{(set_of_face_vertex_indices, element, face_number)}
      It returns a list of tags that apply to this surface.
    @param element_tagger: a function that takes the arguments
      (element) and returns the a list of tags that apply
      to that element.
    @param periodicity: either None or is a list of tuples
      just like the one documented for the C{periodicity}
      member of class L{Mesh}.
    @param _is_rankbdry_face: an implementation detail, 
      should not be used from user code. It is a function
      returning whether a given face identified by 
      C{(element instance, face_nr)} is cut by a parallel
      mesh partition.
    """
    if len(points) == 0:
        raise ValueError, "mesh contains no points"

    dim = len(points[0])
    if dim == 1:
        el_class = Interval
    elif dim == 2:
        el_class = Triangle
    elif dim == 3:
        el_class = Tetrahedron
    else:
        raise ValueError, "%d-dimensional meshes are unsupported" % dim

    # build points and elements
    new_points = numpy.array(points, dtype=float, order="C")

    element_objs = [make_element(el_class, id, vert_indices, new_points) 
        for id, vert_indices in enumerate(elements)]

    # tag elements
    tag_to_elements = {TAG_NONE: [], TAG_ALL: []}
    for el in element_objs:
        for el_tag in element_tagger(el):
            tag_to_elements.setdefault(el_tag, []).append(el)
        tag_to_elements[TAG_ALL].append(el)
    
    # build connectivity
    if periodicity is None:
        periodicity = dim*[None]
    assert len(periodicity) == dim

    mdd = _build_mesh_data_dict(
            new_points, element_objs, boundary_tagger, periodicity, _is_rankbdry_face)
    mdd["tag_to_elements"] = tag_to_elements
    return ConformalMesh(new_points, element_objs, **mdd)




class ConformalMesh(Mesh):
    """A mesh whose elements' faces exactly match up with one another.

    See the Mesh class for data members provided by this class.
    """

    def __init__(self, points, elements, interfaces, tag_to_boundary, tag_to_elements,
            periodicity, periodic_opposite_faces, periodic_opposite_vertices):
        """This constructor is for internal use only. Use L{make_conformal_mesh} instead.
        """
        Mesh.__init__(self, locals())

    def get_reorder_oldnumbers(self, method):
        if method == "cuthill":
            from hedge.tools import cuthill_mckee
            return cuthill_mckee(self.element_adjacency_graph())
        else:
            raise ValueError, "invalid mesh reorder method"

    def reordered_by(self, method):
        old_numbers = self.get_reorder_oldnumbers(method)
        return self.reordered(old_numbers)

    def reordered(self, old_numbers):
        """Return a copy of this C{Mesh} whose elements are 
        reordered using such that for each element C{i},
        C{old_numbers[i]} gives the previous number of that
        element.
        """

        elements = [self.elements[old_numbers[i]].copy(id=i)
                for i in range(len(self.elements))]

        old2new_el = dict(
                (self.elements[old_numbers[i]], new_el) 
                for i, new_el in enumerate(elements)
                )

        # sort interfaces by element id -- this is actually the most important part
        def face_cmp(face1, face2):
            (face1_el1, _), (face1_el2, _) = face1
            (face2_el1, _), (face2_el2, _) = face2

            return cmp(
                    min(face1_el1.id, face1_el2.id), 
                    min(face2_el1.id, face2_el2.id))

        interfaces = [
                ((old2new_el[e1], f1), (old2new_el[e2], f2))
                for (e1,f1), (e2,f2) in self.interfaces]
        interfaces.sort(face_cmp)

        tag_to_boundary = dict(
                (tag, [(old2new_el[old_el], fnr) for old_el, fnr in elfaces])
                for tag, elfaces in self.tag_to_boundary.iteritems())

        tag_to_elements = dict(
                (tag, [old2new_el[old_el] for old_el in tag_els])
                for tag, tag_els in self.tag_to_elements.iteritems())

        return ConformalMesh(
                self.points, elements, interfaces,
                tag_to_boundary, tag_to_elements, self.periodicity,
                self.periodic_opposite_faces, self.periodic_opposite_vertices
                )




def check_bc_coverage(mesh, bc_tags):
    """Verify complete boundary condition coverage.
    
    Given a list of boundary tags as C{bc_tags}, this function verifies
    that
        1. the union of all these boundaries gives the complete boundary,
        2. all these boundaries are disjoint.
    """

    bdry_to_tag = {}
    all_bdry_faces = mesh.tag_to_boundary[TAG_ALL]
    bdry_face_countdown = len(all_bdry_faces)
    bdry_face_countdown_2 = len(set(mesh.tag_to_boundary[TAG_ALL]))

    for tag in bc_tags:
        for el_face in mesh.tag_to_boundary.get(tag, []):
            if el_face in bdry_to_tag:
                raise RuntimeError, "Duplicate BCs %s found on (el=%d,face=%d)" % (
                        [bdry_to_tag[el_face], tag],el_face[0].id, el_face[1])
            else:
                bdry_to_tag[el_face] = tag
                bdry_face_countdown -= 1

    if bdry_face_countdown > 0:
        raise RuntimeError, "No BCs on faces %s" % (
                set(all_bdry_faces)-set(bdry_to_tag.keys()))
    elif bdry_face_countdown < 0:
        raise RuntimeError, "More BCs were assigned than boundary faces are present " \
                "(did something screw up your periodicity?)"




# mesh producers for simple geometries ----------------------------------------
def make_1d_mesh(points, left_tag=None, right_tag=None, periodic=False, 
        boundary_tagger=None):
    def force_array(pt):
        if not isinstance(pt, numpy.ndarray):
            return numpy.array([pt])
        else:
            return pt

    def my_boundary_tagger(fvi, el, fn):
        if el.face_normals[fn][0] < 0:
            return [left_tag]
        else:
            return [right_tag]

    if periodic:
        left_tag = "x_minus"
        right_tag = "x_plus"
        return make_conformal_mesh(
                [force_array(pt) for pt in points],
                [(i,i+1) for i in range(len(points)-1)],
                periodicity=[("x_minus", "x_plus")],
                boundary_tagger=my_boundary_tagger)
    else:
        return make_conformal_mesh(
                [force_array(pt) for pt in points],
                [(i,i+1) for i in range(len(points)-1)],
                boundary_tagger=boundary_tagger or my_boundary_tagger)





def make_uniform_1d_mesh(a, b, el_count, left_tag=None, right_tag=None, periodic=False,
        boundary_tagger=None):
    dx = (b-a)/el_count
    return make_1d_mesh(
            [a+dx*i for i in range(el_count+1)],
            left_tag=left_tag,
            right_tag=right_tag,
            periodic=periodic,
            boundary_tagger=boundary_tagger)




def make_single_element_mesh(a=-0.5, b=0.5, 
        boundary_tagger=(lambda vertices, face_indices: [])):
    n = 2
    node_dict = {}
    points = []
    points_1d = num.linspace(a, b, n)
    for j in range(n):
        for i in range(n):
            node_dict[i,j] = len(points)
            points.append(num.array([points_1d[i], points_1d[j]]))

    elements = [(
                node_dict[1,1],
                node_dict[0,1],
                node_dict[1,0],
                )]

    boundary_faces = [(3,1), (1,2), (2,3)]

    boundary_tags = dict(
            (frozenset(seg), 
                boundary_tagger(points, seg))
                for seg in  boundary_faces)

    return make_conformal_mesh(
            points,
            elements,
            boundary_tags)




def make_regular_rect_mesh(a=(0,0), b=(1,1), n=(5,5), periodicity=None,
        boundary_tagger=(lambda fvi, el, fn: [])):
    """Create a semi-structured rectangular mesh.

    @arg a: the lower left hand point of the rectangle
    @arg b: the upper right hand point of the rectangle
    @arg n: a tuple of integers indicating the total number of points
      on [a,b].
    @arg periodicity: either None, or a tuple of bools specifying whether
      the mesh is to be periodic in x and y.
    """
    node_dict = {}
    points = []
    points_1d = [numpy.linspace(a_i, b_i, n_i)
            for a_i, b_i, n_i in zip(a, b, n)]

    for j in range(n[1]):
        for i in range(n[0]):
            node_dict[i,j] = len(points)
            points.append(numpy.array([points_1d[0][i], points_1d[1][j]]))

    elements = []

    if periodicity is None:
        periodicity = (False, False)

    axes = ["x", "y"]
    mesh_periodicity = []
    periodic_tags = set()
    for i, axis in enumerate(axes):
        if periodicity[i]:
            minus_tag = "minus_"+axis
            plus_tag = "plus_"+axis
            mesh_periodicity.append((minus_tag, plus_tag))
            periodic_tags.add(minus_tag)
            periodic_tags.add(plus_tag)
        else:
            mesh_periodicity.append(None)

    fvi2fm = {}

    for i in range(n[0]-1):
        for j in range(n[1]-1):

            # c--d
            # |  |
            # a--b

            a = node_dict[i,j]
            b = node_dict[i+1,j]
            c = node_dict[i,j+1]
            d = node_dict[i+1,j+1]

            elements.append((a,b,c))
            elements.append((d,c,b))

            if i == 0: fvi2fm[frozenset((a,c))] = "minus_x"
            if i == n[0]-2: fvi2fm[frozenset((b,d))] = "plus_x"
            if j == 0: fvi2fm[frozenset((a,b))] = "minus_y"
            if j == n[1]-2: fvi2fm[frozenset((c,d))] = "plus_y"

    def wrapped_boundary_tagger(fvi, el, fn):
        btag = fvi2fm[frozenset(fvi)]
        if btag in periodic_tags:
            return [btag]
        else:
            return [btag] + boundary_tagger(fvi, el, fn)

    return make_conformal_mesh(points, elements, wrapped_boundary_tagger,
            periodicity=mesh_periodicity)




def make_regular_square_mesh(a=-0.5, b=0.5, n=5, periodicity=None,
        boundary_tagger=(lambda fvi, el, fn: [])):
    """Create a semi-structured square mesh.

    @arg a: the lower x and y coordinate of the square
    @arg b: the upper x and y coordinate of the square
    @arg n: integer indicating the total number of points on [a,b].
    @arg periodicity: either None, or a tuple of bools specifying whether
      the mesh is to be periodic in x and y.
    """
    return make_regular_rect_mesh(
            (a,a), (b,b), (n,n), periodicity, boundary_tagger)




def finish_2d_rect_mesh(points, facets, facet_markers, marker2tag, refine_func,
        periodicity, boundary_tagger):
    """Semi-internal bottom-half routine for generation of rectangular 2D meshes."""
    import meshpy.triangle as triangle

    mesh_info = triangle.MeshInfo()
    mesh_info.set_points(points)
    mesh_info.set_facets(facets, facet_markers)

    #triangle.write_gnuplot_mesh("mesh.dat", mesh_info, True)
    
    if periodicity is None:
        periodicity = (False, False)

    axes = ["x", "y"]
    mesh_periodicity = []
    periodic_tags = set()
    for i, axis in enumerate(axes):
        if periodicity[i]:
            minus_tag = "minus_"+axis
            plus_tag = "plus_"+axis
            mesh_periodicity.append((minus_tag, plus_tag))
            periodic_tags.add(minus_tag)
            periodic_tags.add(plus_tag)
        else:
            mesh_periodicity.append(None)

    generated_mesh = triangle.build(mesh_info, 
            refinement_func=refine_func,
            allow_boundary_steiner=not (periodicity[0] or periodicity[1]))

    fvi2fm = dict((frozenset(fvi), marker) for fvi, marker in
        zip(generated_mesh.facets, generated_mesh.facet_markers))

    def wrapped_boundary_tagger(fvi, el, fn):
        btag = marker2tag[fvi2fm[frozenset(fvi)]]
        if btag in periodic_tags:
            return [btag]
        else:
            return [btag] + boundary_tagger(fvi, el, fn)

    return make_conformal_mesh(
            generated_mesh.points,
            generated_mesh.elements,
            wrapped_boundary_tagger,
            periodicity=mesh_periodicity)




def make_rect_mesh(a=(0,0), b=(1,1), max_area=None, 
        boundary_tagger=(lambda fvi, el, fn: []),
        periodicity=None, subdivisions=None,
        refine_func=None):
    """Create an unstructured rectangular mesh.

    @arg a: the lower left hand point of the rectangle
    @arg b: the upper right hand point of the rectangle
    @arg max_area: maximum area of each triangle.
    @arg periodicity: either None, or a tuple of bools specifying whether
      the mesh is to be periodic in x and y.
    @arg subdivisions: If not C{None}, this is a 2-tuple specifying
      the number of facet subdivisions in X and Y.
    @arg refine_func: A refinement function as taken by C{meshpy.triangle.build}.
    """
    import meshpy.triangle as triangle

    def round_trip_connect(start, end):
        for i in range(start, end):
            yield i, i+1
        yield end, start

    if max_area is not None:
        if refine_func is not None:
            raise ValueError, "cannot specify both refine_func and max_area"
        def refine_func(vertices, area):
            return area > max_area

    marker2tag = {
            1: "minus_x", 
            2: "minus_y", 
            3: "plus_x", 
            4: "plus_y", 
            }

    points = [a, (b[0],a[1]), b, (a[0],b[1])]
    facets = list(round_trip_connect(0, 3))
    facet_markers = [2,3,4,1]

    if subdivisions is not None:
        points, facets, facet_markers = triangle.subdivide_facets(
                [subdivisions[0], subdivisions[1], 
                    subdivisions[0], subdivisions[1]],
                points, facets, facet_markers)
            
    from hedge.mesh import finish_2d_rect_mesh
    return finish_2d_rect_mesh(points, facets, facet_markers, marker2tag, 
            refine_func, periodicity, boundary_tagger)





def make_square_mesh(a=-0.5, b=0.5, max_area=4e-3, 
        boundary_tagger=(lambda fvi, el, fn: [])):
    """Create an unstructured square mesh.

    @arg a: the lower x and y coordinate of the square
    @arg b: the upper x and y coordinate of the square
    @arg max_area: maximum area of each triangle
    """
    return make_rect_mesh((a,a), (b,b), max_area, boundary_tagger)




def make_disk_mesh(r=0.5, faces=50, max_area=4e-3, 
        boundary_tagger=(lambda fvi, el, fn: [])):
    from math import cos, sin, pi

    def round_trip_connect(start, end):
        for i in range(start, end):
            yield i, i+1
        yield end, start

    def needs_refinement(vertices, area):
        return area > max_area

    points = [(r*cos(angle), r*sin(angle))
            for angle in numpy.linspace(0, 2*pi, faces, endpoint=False)]
            
    import meshpy.triangle as triangle

    mesh_info = triangle.MeshInfo()
    mesh_info.set_points(points)
    mesh_info.set_facets(
            list(round_trip_connect(0, faces-1)),
            faces*[1]
            )

    generated_mesh = triangle.build(mesh_info, refinement_func=needs_refinement)

    return make_conformal_mesh(
            generated_mesh.points,
            generated_mesh.elements,
            boundary_tagger)




def make_ball_mesh(r=0.5, subdivisions=10, max_volume=None,
        boundary_tagger=(lambda fvi, el, fn: [])):
    from math import pi, cos, sin
    from meshpy.tet import MeshInfo, build, generate_surface_of_revolution,\
            EXT_OPEN

    dphi = pi/subdivisions

    def truncate(r):
        if abs(r) < 1e-10:
            return 0
        else:
            return r

    rz = [(truncate(r*sin(i*dphi)), r*cos(i*dphi)) for i in range(subdivisions+1)]

    mesh_info = MeshInfo()
    points, facets, facet_holestarts, facet_markers = generate_surface_of_revolution(
            rz, closure=EXT_OPEN, radial_subdiv=subdivisions)

    mesh_info.set_points(points)
    mesh_info.set_facets_ex(facets, facet_holestarts, facet_markers)
    generated_mesh = build(mesh_info, max_volume=max_volume)

    return make_conformal_mesh(
            generated_mesh.points,
            generated_mesh.elements,
            boundary_tagger)




MINUS_X_MARKER = 1
PLUS_X_MARKER = 2
MINUS_Y_MARKER = 3
PLUS_Y_MARKER = 4
MINUS_Z_MARKER = 5
PLUS_Z_MARKER = 6
SHELL_MARKER = 100




def _make_z_periodic_mesh(points, facets, facet_holestarts, facet_markers, height, 
        max_volume, boundary_tagger):
    from meshpy.tet import MeshInfo, build

    mesh_info = MeshInfo()
    mesh_info.set_points(points)
    mesh_info.set_facets_ex(facets, facet_holestarts, facet_markers)

    mesh_info.pbc_groups.resize(1)
    pbcg = mesh_info.pbc_groups[0]

    pbcg.facet_marker_1 = MINUS_Z_MARKER
    pbcg.facet_marker_2 = PLUS_Z_MARKER

    pbcg.set_transform(translation=[0,0,height])

    def zper_boundary_tagger(fvi, el, fn):
        # we only ask about *boundaries*
        # we should not try to have the user tag
        # the (periodicity-induced) interior faces

        face_marker = fvi2fm[frozenset(fvi)]

        if face_marker == MINUS_Z_MARKER:
            return ["minus_z"]
        if face_marker == PLUS_Z_MARKER:
            return ["plus_z"]

        result = boundary_tagger(fvi, el, fn)
        if face_marker == SHELL_MARKER:
            result.append("shell")
        return result

    generated_mesh = build(mesh_info, max_volume=max_volume)
    fvi2fm = generated_mesh.face_vertex_indices_to_face_marker
        
    return make_conformal_mesh(
            generated_mesh.points,
            generated_mesh.elements,
            zper_boundary_tagger,
            periodicity=[None, None, ("minus_z", "plus_z")])




def make_cylinder_mesh(radius=0.5, height=1, radial_subdivisions=10, 
        height_subdivisions=1, max_volume=None, periodic=False,
        boundary_tagger=(lambda fvi, el, fn: [])):
    from math import pi, cos, sin
    from meshpy.tet import MeshInfo, build, generate_surface_of_revolution, \
            EXT_OPEN

    dz = height/height_subdivisions
    rz = [(0,0)] \
            + [(radius, i*dz) for i in range(height_subdivisions+1)] \
            + [(0,height)]
    ring_markers = [MINUS_Z_MARKER] \
            + ((height_subdivisions)*[SHELL_MARKER]) \
            + [PLUS_Z_MARKER]

    points, facets, facet_holestarts, facet_markers = generate_surface_of_revolution(rz,
            closure=EXT_OPEN, radial_subdiv=radial_subdivisions,
            ring_markers=ring_markers)

    assert len(facets) == len(facet_markers)

    if periodic:
        return _make_z_periodic_mesh(
                points, facets, facet_holestarts, facet_markers, 
                height=height, 
                max_volume=max_volume,
                boundary_tagger=boundary_tagger)
    else:
        mesh_info = MeshInfo()
        mesh_info.set_points(points)
        mesh_info.set_facets_ex(facets, facet_holestarts, facet_markers)

        generated_mesh = build(mesh_info, max_volume=max_volume)

        return make_conformal_mesh(
                generated_mesh.points,
                generated_mesh.elements,
                boundary_tagger)




def make_box_mesh(dimensions=(1,1,1), max_volume=None, periodicity=None,
        boundary_tagger=(lambda fvi, el, fn: [])):
    """Return a mesh for a brick from the origin to `dimensions'.

    `max_volume' specifies the maximum volume for each tetrahedron.
    `periodicity' is either None, or a triple of bools, indicating
    whether periodic BCs are to be applied along that axis.
    See ConformalMesh.__init__ for the meaning of boundary_tagger.

    A few stock boundary tags are provided for easy application
    of boundary conditions, namely plus_[xyz] and minus_[xyz] tag
    the appropriate faces of the brick.
    """

    def count(iterable):
        result = 0
        for i in iterable:
            result += 1
        return result

    from meshpy.tet import MeshInfo, build

    d = dimensions

    #    7--------6
    #   /|       /|
    #  4--------5 |  z
    #  | |      | |  ^
    #  | 3------|-2  | y
    #  |/       |/   |/
    #  0--------1    +--->x

    points = [
            (0   ,   0,   0),
            (d[0],   0,   0),
            (d[0],d[1],   0),
            (0   ,d[1],   0),
            (0   ,   0,d[2]),
            (d[0],   0,d[2]),
            (d[0],d[1],d[2]),
            (0   ,d[1],d[2]),
            ]

    facets = [
            (0,1,2,3),
            (0,1,5,4),
            (1,2,6,5),
            (7,6,2,3),
            (7,3,0,4),
            (4,5,6,7)
            ]

    tags = [MINUS_Z_MARKER, MINUS_Y_MARKER, PLUS_X_MARKER, 
            PLUS_Y_MARKER, MINUS_X_MARKER, PLUS_Z_MARKER]
            
    mesh_info = MeshInfo()
    mesh_info.set_points(points)
    mesh_info.set_facets(facets, tags)

    if periodicity is None:
        periodicity = (False, False, False)

    axes = ["x", "y", "z"]

    per_count = count(p for p in periodicity if p)
    mesh_info.pbc_groups.resize(per_count)
    pbc_group_number = 0

    marker_to_tag = {}
    mesh_periodicity = []
    periodic_tags = set()

    for axis, axis_per in enumerate(periodicity):
        minus_marker = 1+2*axis
        plus_marker = 2+2*axis

        minus_tag = "minus_"+axes[axis]
        plus_tag = "plus_"+axes[axis]

        marker_to_tag[minus_marker] = minus_tag
        marker_to_tag[plus_marker] = plus_tag

        if axis_per:
            pbcg = mesh_info.pbc_groups[pbc_group_number]
            pbc_group_number +=1

            pbcg.facet_marker_1 = minus_marker
            pbcg.facet_marker_2 = plus_marker

            translation = [0,0,0]
            translation[axis] = d[axis]
            pbcg.set_transform(translation=translation)

            mesh_periodicity.append((minus_tag, plus_tag))
            periodic_tags.add(minus_tag)
            periodic_tags.add(plus_tag)
        else:
            mesh_periodicity.append(None)

    generated_mesh = build(mesh_info, max_volume=max_volume)

    fvi2fm = generated_mesh.face_vertex_indices_to_face_marker

    def wrapped_boundary_tagger(fvi, el, fn):
        face_tag = marker_to_tag[fvi2fm[frozenset(fvi)]]

        if face_tag in periodic_tags:
            return [face_tag]
        else:
            return [face_tag] + boundary_tagger(fvi, el, fn)

    return make_conformal_mesh(
            generated_mesh.points,
            generated_mesh.elements,
            wrapped_boundary_tagger,
            periodicity=mesh_periodicity)




# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




from __future__ import division
import numpy
import numpy.linalg as la
from hedge.tools import Reflection, Rotation




def main() :
    from hedge.element import \
            TriangularElement, \
            TetrahedralElement
    from hedge.operators import WeakPoissonOperator
    from hedge.mesh import \
            make_disk_mesh, \
            make_regular_square_mesh, \
            make_square_mesh, \
            make_ball_mesh, \
            TAG_ALL, TAG_NONE
    from hedge.visualization import SiloVisualizer, VtkVisualizer
    from math import sin, cos, pi, exp, sqrt
    from hedge.parallel import guess_parallelization_context
    from hedge.data import GivenFunction, ConstantGivenFunction

    pcon = guess_parallelization_context()

    dim = 2

    def boundary_tagger(fvi, el, fn):
        from math import atan2, pi
        normal = el.face_normals[fn]
        if -10/180*pi < atan2(normal[1], normal[0]) < 10/180*pi:
            return ["dirichlet"]
        else:
            return ["neumann"]

    def boundary_tagger_2(fvi, el, fn):
        if el.face_normals[fn][0] > 0:
            return ["dirichlet"]
        else:
            return ["neumann"]

    if dim == 2:
        if pcon.is_head_rank:
            mesh = make_disk_mesh(r=0.5, boundary_tagger=boundary_tagger,
                    max_area=1e-2)
            #mesh = make_regular_square_mesh(n=3, boundary_tagger=boundary_tagger, periodicity=(True,False))
            #mesh = make_regular_square_mesh(n=9)
            #mesh = make_square_mesh(max_area=0.1, boundary_tagger=boundary_tagger)
            #mesh.transform(Reflection(0,2))
        el_class = TriangularElement
    elif dim == 3:
        if pcon.is_head_rank:
            mesh = make_ball_mesh(max_volume=0.0001)
        el_class = TetrahedralElement
    else:
        raise RuntimeError, "bad number of dimensions"

    if pcon.is_head_rank:
        print "%d elements" % len(mesh.elements)
        mesh_data = pcon.distribute_mesh(mesh)
    else:
        mesh_data = pcon.receive_mesh()

    discr = pcon.make_discretization(mesh_data, el_class(5))
    vis = VtkVisualizer(discr, pcon)

    def u0(x):
        if comp.norm_2(x) < 0.2:
            #return exp(-100*x*x)
            return 1
        else:
            return 0

    def dirichlet_bc(x):
        if x[0] > 0:
            return 0
        else:
            return 1

    def rhs_c(x):
        if la.norm(x) < 0.1:
            return 1000
        else:
            return 0

    class DiffTensor:
        shape = (dim, dim)

        def __call__(self, x):
            result = num.identity(dim)
            result[0,0] = 20*abs(x[0]-0.5)
            return result

    def my_diff_tensor():
        result = numpy.eye(dim)
        result[0,0] = 0.1
        return result

    op = WeakPoissonOperator(discr, 
            diffusion_tensor=ConstantGivenFunction(my_diff_tensor()),
            #diffusion_tensor=GivenFunction(DiffTensor()),

            #dirichlet_tag="dirichlet",
            #neumann_tag="neumann", 

            dirichlet_tag=TAG_ALL,
            neumann_tag=TAG_NONE,

            #dirichlet_bc=GivenFunction(dirichlet_bc),
            dirichlet_bc=ConstantGivenFunction(0),
            neumann_bc=ConstantGivenFunction(-10),
            )

    def l2_norm(v):
        return sqrt(v*(discr.mass_operator*v))

    def matrix_rep(op):
        h,w = op.shape
        mat = num.zeros(op.shape)
        for j in range(w):
            mat[:,j] = op(num.unit_vector(w, j))
        return mat

    if False:
        mat = matrix_rep(op)
        print comp.norm_frobenius(mat-mat.T)
        #print comp.eigenvalues(mat)
        print mat.shape
    
    if False:
        from hedge.discretization import PylinearOpWrapper
        results = comp.operator_eigenvectors(-op, 30, PylinearOpWrapper(discr.mass_operator),
                which=comp.SMALLEST_MAGNITUDE
                )
        scalars = []
        for i, (value,vector) in enumerate(results):
            print i, value, l2_norm(vector.real)
            scalars.append(("ev%03d" % i, vector.real))
        print 

        visf = vis.make_file("eigenvectors")
        vis.add_data(visf, scalars)
        visf.close()
        return


    #a_inv = operator.BiCGSTABOperator.make(-op, 40000, 1e-10)
    #a_inv = operator.CGOperator.make(-op, 40000, 1e-4)
    #a_inv.debug_level = 1
    #u = -a_inv(op.prepare_rhs(GivenFunction(rhs_c)))

    numpy.seterr('raise')
    if False:
        u = GivenFunction(rhs_c).volume_interpolant(discr)

        N = 30000
        from time import time
        start = time()
        
        #z = num.arange(17)
        #u = numpy.array([z,z])
        #u = ArithmeticList([z,z])
        #u = [z,z]
        for i in xrange(N):
            op.op(u)
            #op.grad(u)

        print (time()-start)/(N)
        return

    from hedge.tools import parallel_cg
    u = -parallel_cg(pcon, -op, op.prepare_rhs(GivenFunction(rhs_c)), 
            debug=True, tol=1e-10)
    print len(u)

    from hedge.discretization import ones_on_boundary
    visf = vis.make_file("fld")
    vis.add_data(visf, [
        ("sol", u), 
        ("dir", ones_on_boundary(discr, "dirichlet")), 
        ("neu", ones_on_boundary(discr, "neumann")), 
        ])
    visf.close()





if __name__ == "__main__":
    #import cProfile as profile
    #profile.run("main()", "poisson.prof")
    main()




"""Representations for given data, such as initial and boundary 
conditions and source terms."""

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import hedge.mesh




# helpers ---------------------------------------------------------------------
class _ConstantFunctionContainer:
    def __init__(self, value):
        self.value = value

    @property
    def shape(self):
        return self.value.shape

    def __call__(self, x):
        return self.value





# interpolation wrappers ------------------------------------------------------
class IGivenFunction(object):
    """Abstract interface for obtaining interpolants of I{time-independent} 
    functions.
    """

    def volume_interpolant(self, discr):
        """Return the volume interpolant of this function with respect to
        the L{Discretization} C{discr}.
        """
        raise NotImplementedError

    def boundary_interpolant(self, discr, tag=hedge.mesh.TAG_ALL):
        """Return the boundary interpolant of this function with respect to
        the L{Discretization} discr at the boundary tagged with C{tag}.
        """
        raise NotImplementedError




class ITimeDependentGivenFunction(object):
    """Abstract interface for obtaining interpolants of I{time-dependent} 
    functions.
    """

    def volume_interpolant(self, t, discr):
        """Return the volume interpolant of this function with respect to
        the L{Discretization} discr at time {t}.
        """
        raise NotImplementedError

    def boundary_interpolant(self, t, discr, tag=hedge.mesh.TAG_ALL):
        """Return the boundary interpolant of this function with respect to
        the L{Discretization} discr at time C{t} at the boundary tagged with
        C{tag}.
        """
        raise NotImplementedError




class GivenFunction(IGivenFunction):
    """Adapter for a function M{f(x)} into an L{IGivenFunction}.
    """
    def __init__(self, f):
        """Initialize the caches and store the function C{f}.

        @param f: a function mapping space to a scalar value.
          If f.target_dimensions exists and equals M{n}, then f maps into an
          M{n}-dimensional vector space instead.
        """
        from weakref import WeakKeyDictionary

        self.f = f

        self.volume_cache = WeakKeyDictionary()
        self.boundary_cache = WeakKeyDictionary()

    def volume_interpolant(self, discr):
        try:
            return self.volume_cache[discr]
        except KeyError:
            result = discr.interpolate_volume_function(self.f)
            self.volume_cache[discr] = result
            return result

    def boundary_interpolant(self, discr, tag=hedge.mesh.TAG_ALL):
        try:
            return self.boundary_cache[discr][tag]
        except KeyError:
            tag_cache = self.boundary_cache.setdefault(discr, {})
            result = discr.interpolate_boundary_function(self.f, tag)
            tag_cache[tag] = result
            return result




class ConstantGivenFunction(GivenFunction):
    """A L{GivenFunction} that has a constant value on all space.
    """
    def __init__(self, value=0):
        self.value = value

        GivenFunction.__init__(self, _ConstantFunctionContainer(value))




class GivenVolumeInterpolant(IGivenFunction):
    """A constant-valued L{GivenFunction}.
    """
    def __init__(self, discr, interpolant):
        self.discr = discr
        self.interpolant = interpolant

    def volume_interpolant(self, discr):
        if discr != self.discr:
            raise ValueError, "cross-interpolation between discretizations not supported"
        return self.interpolant

    def boundary_interpolant(self, discr, tag=hedge.mesh.TAG_ALL):
        if discr != self.discr:
            raise ValueError, "cross-interpolation between discretizations not supported"
        return discr.boundarize_volume_field(self.interpolant, tag)





class TimeConstantGivenFunction(ITimeDependentGivenFunction):
    """Adapts a L{GivenFunction} to have a (formal) time-dependency, being constant
    over all time.
    """
    def __init__(self, gf):
        self.gf = gf

    def volume_interpolant(self, t, discr):
        return self.gf.volume_interpolant(discr)

    def boundary_interpolant(self, t, discr, tag=hedge.mesh.TAG_ALL):
        return self.gf.boundary_interpolant(discr, tag)





def make_tdep_constant(x):
    return TimeConstantGivenFunction(ConstantGivenFunction(x))




class TimeHarmonicGivenFunction(ITimeDependentGivenFunction):
    """Adapts an L{IGivenFunction} to have a harmonic time-dependency.
    """
    def __init__(self, gf, omega, phase=0):
        self.gf = gf
        self.omega = omega
        self.phase = phase

    def volume_interpolant(self, t, discr):
        from math import sin
        return sin(omega*t+phase)*self.gf.volume_interpolant(discr)

    def boundary_interpolant(self, t, discr, tag=hedge.mesh.TAG_ALL):
        from math import sin
        return sin(omega*t+phase)*self.gf.boundary_interpolant(discr, tag)





class TimeIntervalGivenFunction(ITimeDependentGivenFunction):
    """Adapts an L{IGivenFunction} to depend on time by "turning it on"
    for the time interval [on_time, off_time), and having it be zero
    the rest of the time.
    """

    def __init__(self, gf, on_time=0, off_time=1):
        self.gf = gf
        self.on_time = on_time
        self.off_time = off_time
        assert on_time <= off_time



    def volume_interpolant(self, t, discr):
        if self.on_time <= t < self.off_time:
            return self.gf.volume_interpolant(discr)
        else:
            # FIXME: not optimal
            # difficult part here is to match shape
            return 0*self.gf.volume_interpolant(discr)

    def boundary_interpolant(self, t, discr, tag=hedge.mesh.TAG_ALL):
        if self.on_time <= t < self.off_time:
            return self.gf.boundary_interpolant(discr, tag)
        else:
            # FIXME: not optimal
            # difficult part here is to match shape
            return 0*self.gf.boundary_interpolant(discr, tag)





class TimeDependentGivenFunction(ITimeDependentGivenFunction):
    """Adapts a function M{f(x,t)} into the L{GivenFunction} framework.
    """
    def __init__(self, f):
        self.f = f

    def volume_interpolant(self, t, discr):
        return discr.interpolate_volume_function(lambda x: self.f(x,t))

    def boundary_interpolant(self, t, discr, tag=hedge.mesh.TAG_ALL):
        return discr.interpolate_boundary_function(lambda x: self.f(x, t), tag)



        from hedge.discretization import ones_on_boundary
        bdry1 = ones_on_boundary(discr, None)
        silo = SiloFile("bdry.silo")
        vis.add_to_silo(silo,
                scalars=[("bdry", bdry1)],
                write_coarse_mesh=True,
                )
        return

        from hedge.visualization import SiloVisualizer
        from hedge.silo import SiloFile, DB_CLOBBER
        vis = SiloVisualizer(self.discr)
        silo = SiloFile("flux.silo", mode=DB_CLOBBER)
        vis.add_to_silo(silo,
                vectors=[
                    ("int_h", cross(self.n_jump, h)),
                    ("bdry_h", cross(self.n_jump, h_pair)),
                    ],
                write_coarse_mesh=True,
                )
        silo.close()
        raw_input("Enter:")




"""Building blocks for flux computation. Flux compilation."""

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import numpy
import hedge._internal as _internal
from pytools import monkeypatch_method
import pymbolic
import pymbolic.mapper.collector




FluxFace = _internal.FluxFace




# c++ fluxes debug dumping ----------------------------------------------------
@monkeypatch_method(_internal.ConstantFlux)
def __str__(self):
    return "ConstantFlux(%s)" % self.value
@monkeypatch_method(_internal.NormalFlux)
def __str__(self):
    return "NormalFlux(%d)" % self.axis
@monkeypatch_method(_internal.PenaltyFlux)
def __str__(self):
    return "PenaltyFlux(%s)" % self.power
@monkeypatch_method(_internal.SumFlux)
def __str__(self):
    return "SumFlux(%s, %s)" % (self.operand1, self.operand2)
@monkeypatch_method(_internal.ProductFlux)
def __str__(self):
    return "ProductFlux(%s, %s)" % (self.operand1, self.operand2)
@monkeypatch_method(_internal.NegativeFlux)
def __str__(self):
    return "NegativeFlux(%s)" % self.operand
@monkeypatch_method(_internal.ChainedFlux)
def __str__(self):
    #return "ChainedFlux(%s)" % self.child
    return str(self.child)





# python fluxes ---------------------------------------------------------------
class Flux(pymbolic.primitives.AlgebraicLeaf, _internal.Flux):
    def stringify(self, enclosing_prec, use_repr_for_constants=False):
        return FluxStringifyMapper(use_repr_for_constants)(self, enclosing_prec)

    def perform(self, face_group, which_faces, fmm, target):
        _internal.ChainedFlux(self).perform(face_group, which_faces, fmm, target)




class FieldComponent(Flux):
    def __init__(self, index, is_local):
        self.index = index
        self.is_local = is_local

    def __eq__(self, other):
        return (isinstance(other, FieldComponent) 
                and self.index == other.index
                and self.is_local == other.is_local
                )

    def __getinitargs__(self):
        return self.index, self.is_local

    def __hash__(self):
        return 0x7371afcd ^ hash(self.index) ^ hash(self.is_local)

    def get_mapper_method(self, mapper):
        return mapper.map_field_component




class Normal(Flux):
    def __init__(self, axis):
        self.axis = axis

    def __getinitargs__(self):
        return self.axis,

    def __eq__(self, other):
        return isinstance(other, Normal) and self.axis == other.axis

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return 0xa931ff ^ hash(self.axis)

    def get_mapper_method(self, mapper):
        return mapper.map_normal




class PenaltyTerm(Flux):
    def __init__(self, power=1):
        self.power = power

    def __getinitargs__(self):
        return self.power,

    def __eq__(self, other):
        return isinstance(other, PenaltyTerm) and self.power == other.power

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return 0x9f194 ^ hash(self.power)

    def get_mapper_method(self, mapper):
        return mapper.map_penalty_term




class IfPositive(Flux):
    def __init__(self, criterion, then, else_):
        self.criterion = criterion
        self.then = then
        self.else_ = else_

        if FluxDependencyMapper(composite_leaves=True)(criterion):
            raise ValueError("criterion part of IfPositive may not depend on field values")

    def __getinitargs__(self):
        return self.criterion, self.then, self.else_

    def __eq__(self, other):
        return (isinstance(other, IfPositive)
                and self.criterion == other.criterion
                and self.then == other.then
                and self.else_ == other.else_)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return 0xfa5814 ^ hash(self.criterion) ^ hash(self.then) ^ hash(self.else_)

    def get_mapper_method(self, mapper):
        return mapper.map_if_positive




def make_normal(dimensions):
    from hedge.tools import amap
    return amap(Normal,  range(dimensions))





class FluxZeroPlaceholder(object):
    @property
    def int(self):
        return 0

    @property
    def ext(self):
        return 0

    @property
    def avg(self):
        return 0




class FluxScalarPlaceholder(object):
    def __init__(self, component=0):
        self.component = component

    @property
    def int(self):
        return FieldComponent(self.component, True)

    @property
    def ext(self):
        return FieldComponent(self.component, False)

    @property
    def avg(self):
        return 0.5*(self.int+self.ext)




class FluxVectorPlaceholder(object):
    def __init__(self, components=None, scalars=None):
        if not (components or scalars):
            raise ValueError, "either components or scalars must be specified"
        if components and scalars:
            raise ValueError, "only one of components and scalars may be specified"

        if components:
            self.scalars = [
                    FluxScalarPlaceholder(i) 
                    for i in range(components)]
        else:
            self.scalars = scalars

    def __len__(self):
        return len(self.scalars)

    def __getitem__(self, idx):
        if isinstance(idx, int):
            return self.scalars[idx]
        else:
            return FluxVectorPlaceholder(scalars=self.scalars.__getitem__(idx))

    @property
    def int(self):
        return numpy.array([scalar.int for scalar in self.scalars])

    @property
    def ext(self):
        return numpy.array([scalar.ext for scalar in self.scalars])

    @property
    def avg(self):
        return numpy.array([scalar.avg for scalar in self.scalars])




# internal flux wrangling -----------------------------------------------------
class FluxStringifyMapper(pymbolic.mapper.stringifier.StringifyMapper):
    def map_field_component(self, expr, enclosing_prec):
        if expr.is_local:
            return "Int[%d]" % expr.index
        else:
            return "Ext[%d]" % expr.index

    def map_normal(self, expr, enclosing_prec):
        return "Normal(%d)" % expr.axis

    def map_penalty_term(self, expr, enclosing_prec):
        return "Penalty(%s)" % (expr.power)

    def map_if_positive(self, expr, enclosing_prec):
        return "IfPositive(%s, %s, %s)" % (expr.criterion, expr.then, expr.else_)




class FluxNormalizationMapper(pymbolic.mapper.collector.TermCollector):
    def map_constant_flux(self, expr):
        if expr.local_c == expr.neighbor_c:
            return expr.local_c
        else:
            return expr

    def map_if_positive(self, expr):
        return IfPositive(
                self.rec(expr.criterion),
                self.rec(expr.then),
                self.rec(expr.else_),
                )



        
def normalize_flux(flux):
    from pymbolic import expand, flatten
    from pymbolic.mapper.constant_folder import CommutativeConstantFoldingMapper
    return CommutativeConstantFoldingMapper()(FluxNormalizationMapper()(
        flatten(expand(flux))))




class FluxDependencyMapper(pymbolic.mapper.dependency.DependencyMapper):
    def map_field_component(self, expr):
        return set([expr])

    def map_normal(self, expr):
        return set()

    def map_penalty_term(self, expr):
        return set()

    def map_if_positive(self, expr):
        return self.rec(expr.criterion) | self.rec(expr.then) | self.rec(expr.else_)




class FluxCompilationMapper(pymbolic.mapper.RecursiveMapper):
    def handle_unsupported_expression(self, expr):
        if isinstance(expr, _internal.Flux):
            return expr
        else:
            pymbolic.mapper.RecursiveMapper.\
                    handle_unsupported_expression(self, expr)

    def map_constant(self, expr):
        return _internal.ConstantFlux(expr)

    def map_sum(self, expr):
        return reduce(lambda f1, f2: _internal.SumFlux(
                    _internal.ChainedFlux(f1), 
                    _internal.ChainedFlux(f2)),
                (self.rec(c) for c in expr.children))

    def map_product(self, expr):
        return reduce(
                lambda f1, f2: _internal.ProductFlux(
                    _internal.ChainedFlux(f1), 
                    _internal.ChainedFlux(f2)),
                (self.rec(c) for c in expr.children))

    def map_negation(self, expr):
        return _internal.NegativeFlux(_internal.ChainedFlux(self.rec(expr.child)))

    def map_power(self, expr):
        base = self.rec(expr.base)
        result = base

        chain_base = _internal.ChainedFlux(base)

        assert isinstance(expr.exponent, int)

        for i in range(1, expr.exponent):
            result = _internal.ProductFlux(_internal.ChainedFlux(result), chain_base)

        return result

    def map_normal(self, expr):
        return _internal.NormalFlux(expr.axis)

    def map_penalty_term(self, expr):
        return _internal.PenaltyFlux(expr.power)

    def map_if_positive(self, expr):
        return _internal.IfPositiveFlux(
                _internal.ChainedFlux(self.rec(expr.criterion)),
                _internal.ChainedFlux(self.rec(expr.then)),
                _internal.ChainedFlux(self.rec(expr.else_)),
                )




class FluxDifferentiationMapper(pymbolic.mapper.differentiator.DifferentiationMapper):
    def map_field_component(self, expr):
        if expr == self.variable:
            return 1
        else:
            return 0

    def map_normal(self, expr):
        return 0

    def map_penalty_term(self, expr):
        return 0

    def map_if_positive(self, expr):
        return IfPositive(
                expr.criterion,
                self.rec(expr.then),
                self.rec(expr.else_),
                )




def compile_flux(flux):
    from hedge.tools import is_obj_array
    if is_obj_array(flux):
        return numpy.array([compile_flux(subflux) for subflux in flux])

    def compile_scalar_single_dep_flux(flux):
        return FluxCompilationMapper()(flux)

    def compile_scalar_flux(flux):
        def in_fields_cmp(a, b):
            return cmp(a.index, b.index) \
                    or cmp(a.is_local, b.is_local)

        in_fields = list(FluxDependencyMapper(composite_leaves=True)(flux))

        # check that all in_fields are FieldComponents
        for in_field in in_fields:
            if not isinstance(in_field, FieldComponent):
                raise ValueError, "flux depends on invalid term `%s'" % str(in_field)
            
        in_fields.sort(in_fields_cmp)

        result = []

        if in_fields:
            max_in_field = max(in_field.index for in_field in in_fields)

            # find d<flux> / d<in_fields>
            in_derivatives = dict(
                    ((in_field.index, in_field.is_local),
                    normalize_flux(FluxDifferentiationMapper(in_field)(flux)))
                    for in_field in in_fields)

            # check for (invalid) nonlinearity
            for i, deriv in in_derivatives.iteritems():
                if FluxDependencyMapper()(deriv):
                    raise ValueError, "Flux is nonlinear in component %d" % i

            for in_field_idx in range(max_in_field+1):
                int = in_derivatives.get((in_field_idx, True), 0)
                ext = in_derivatives.get((in_field_idx, False), 0)

                if int or ext:
                    result.append(
                            (in_field_idx, 
                                compile_scalar_single_dep_flux(int),
                                compile_scalar_single_dep_flux(ext),))

        try:
            flux._compiled = result
        except AttributeError:
            pass # ok if we can't set cache

        return result

    try:
        return flux._compiled
    except AttributeError:
        return compile_scalar_flux(flux)



# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division




import numpy
import numpy.linalg as la
import unittest




class Monomial:
    def __init__(self, exponents, factor=1):
        self.exponents = exponents
        self.ones = numpy.ones((len(self.exponents),))
        self.factor = factor

    def __call__(self, x):
        from operator import mul

        eps = 1e-15
        x = (x+self.ones)/2
        for xi in x:
            assert -eps <= xi <= 1+eps
        return self.factor* \
                reduce(mul, (x[i]**alpha 
                    for i, alpha in enumerate(self.exponents)))

    def theoretical_integral(self):
        from pytools import factorial
        from operator import mul

        return (self.factor*2**len(self.exponents)*
            reduce(mul, (factorial(alpha) for alpha in self.exponents))
            /
            factorial(len(self.exponents)+sum(self.exponents)))

    def diff(self, coordinate):
        diff_exp = list(self.exponents)
        orig_exp = diff_exp[coordinate]
        if orig_exp == 0:
            return Monomial(diff_exp, 0)
        diff_exp[coordinate] = orig_exp-1
        return Monomial(diff_exp, self.factor*orig_exp)




class TestHedge(unittest.TestCase):
    def test_timestep_accuracy(self):
        """Check that all timesteppers have the advertised accuracy"""
        from math import sqrt, log, sin, cos
        from hedge.tools import EOCRecorder

        def rhs(t, y):
            u = y[0]
            v = y[1]
            rhs = numpy.array([v, -u/t**2])
            return rhs

        def soln(t):
            inner = sqrt(3)/2*log(t)
            return sqrt(t)*(
                    5*sqrt(3)/3*sin(inner)
                    + cos(inner)
                    )

        def get_error(stepper, dt):
            t = 1
            y = numpy.array([1, 3])
            final_t = 10
            nsteps = int((final_t-t)/dt)

            hist = []
            for i in range(nsteps):
                y = stepper(y, t, dt, rhs)
                t += dt
                hist.append(y)

            return abs(y[0]-soln(t))

        def verify_timestep_order(stepper_getter, order):
            eocrec = EOCRecorder()
            for n in range(4,8):
                dt = 2**(-n)
                stepper = stepper_getter()
                error = get_error(stepper,dt)
                eocrec.add_data_point(1/dt, error)

            #print stepper
            #print eocrec.pretty_print()

            orderest = eocrec.estimate_order_of_convergence()[0,1]
            #print orderest, order
            self.assert_(orderest > order*0.95)

        from hedge.timestep import RK4TimeStepper, AdamsBashforthTimeStepper

        for o in range(1,5):
            verify_timestep_order(lambda : AdamsBashforthTimeStepper(o), o)
        verify_timestep_order(RK4TimeStepper, 4)

    # -------------------------------------------------------------------------
    def test_face_vertex_order(self):
        """Verify that face_indices() emits face vertex indices in the right order"""
        from hedge.element import \
                IntervalElement, \
                TriangularElement, \
                TetrahedralElement

        for el in [
                IntervalElement(5),
                TriangularElement(5), 
                TetrahedralElement(5)]:
            vertex_indices = el.vertex_indices()
            for fn, (face_vertices, face_indices) in enumerate(zip(
                    el.geometry.face_vertices(vertex_indices), 
                    el.face_indices())):
                face_vertices_i = 0
                for fi in face_indices:
                    if fi == face_vertices[face_vertices_i]:
                        face_vertices_i += 1

                self.assert_(face_vertices_i == len(face_vertices))

    # -------------------------------------------------------------------------
    def test_newton_interpolation(self):
        """Verify Newton interpolation"""
        from hedge.interpolation import newton_interpolation_function
        
        x = [-1.5, -0.75, 0, 0.75, 1.5]
        y = [-14.1014, -0.931596, 0, 0.931596, 14.1014]
        nf = newton_interpolation_function(x, y)

        errors = [abs(yi-nf(xi)) for xi, yi in zip(x, y)]
        #print errors
        self.assert_(sum(errors) < 1e-14)
    # -------------------------------------------------------------------------
    def test_orthonormality_jacobi_1d(self):
        """Verify that the Jacobi polymials are orthogonal in 1D"""
        from hedge.polynomial import JacobiFunction
        from hedge.quadrature import LegendreGaussQuadrature

        max_n = 10
        int = LegendreGaussQuadrature(4*max_n) # overkill...

        class WeightFunction:
            def __init__(self, alpha, beta):
                self.alpha = alpha
                self.beta = beta

            def __call__(self, x):
                return (1-x)**self.alpha * (1+x)**self.beta

        for alpha, beta, ebound in [
                (0, 0, 5e-14), 
                (1, 0, 4e-14), 
                (3, 2, 3e-14), 
                (0, 2, 3e-13), 
                (5, 0, 3e-13), 
                (3, 4, 1e-14)
                ]:
            jac_f = [JacobiFunction(alpha, beta, n) for n in range(max_n)]
            wf = WeightFunction(alpha, beta)
            maxerr = 0

            for i, fi in enumerate(jac_f):
                for j, fj in enumerate(jac_f):
                    result = int(lambda x: wf(x)*fi(x)*fj(x))

                    if i == j:
                        true_result = 1
                    else:
                        true_result = 0
                    err = abs(result-true_result)
                    maxerr = max(maxerr, err)
                    if abs(result-true_result) > ebound:
                        print "bad", alpha, beta, i, j, abs(result-true_result)
                    self.assert_(abs(result-true_result) < ebound)
            #print alpha, beta, maxerr
    # -------------------------------------------------------------------------
    def test_transformed_quadrature(self):
        """Test 1D quadrature on arbitrary intervals"""
        from math import exp, sqrt, pi

        def gaussian_density(x, mu, sigma):
            return 1/(sigma*sqrt(2*pi))*exp(-(x-mu)**2/(2*sigma**2))

        from hedge.quadrature import LegendreGaussQuadrature, TransformedQuadrature

        mu = 17
        sigma = 12
        tq = TransformedQuadrature(LegendreGaussQuadrature(20), mu-6*sigma, mu+6*sigma)
        
        result = tq(lambda x: gaussian_density(x, mu, sigma))
        self.assert_(abs(result - 1) < 1e-9)
    # -------------------------------------------------------------------------
    def test_warp(self):
        """Check some assumptions on the node warp factor calculator"""
        n = 17
        from hedge.element import WarpFactorCalculator
        wfc = WarpFactorCalculator(n)

        self.assert_(abs(wfc.int_f(-1)) < 1e-12)
        self.assert_(abs(wfc.int_f(1)) < 1e-12)

        from hedge.quadrature import LegendreGaussQuadrature

        lgq = LegendreGaussQuadrature(n)
        self.assert_(abs(lgq(wfc)) < 6e-14)
    # -------------------------------------------------------------------------
    def test_simp_nodes(self):
        """Verify basic assumptions on simplex interpolation nodes"""
        from hedge.element import \
                IntervalElement, \
                TriangularElement, \
                TetrahedralElement

        els = [
                IntervalElement(19), 
                TriangularElement(8), 
                TriangularElement(17), 
                TetrahedralElement(13)]

        for el in els:
            eps = 1e-10

            unodes = list(el.unit_nodes())
            self.assert_(len(unodes) == el.node_count())
            for ux in unodes:
                for uc in ux:
                    self.assert_(uc >= -1-eps)
                self.assert_(sum(ux) <= 1+eps)

            try:
                equnodes = list(el.equidistant_unit_nodes())
            except AttributeError:
                assert isinstance(el, IntervalElement)
            else:
                self.assert_(len(equnodes) == el.node_count())
                for ux in equnodes:
                    for uc in ux:
                        self.assert_(uc >= -1-eps)
                    self.assert_(sum(ux) <= 1+eps)

            for indices in el.node_tuples():
                for index in indices:
                    self.assert_(index >= 0)
                self.assert_(sum(indices) <= el.order)
    # -------------------------------------------------------------------------
    def test_tri_nodes_against_known_values(self):
        """Check triangle nodes against a previous implementation"""
        from hedge.element import TriangularElement, TetrahedralElement
        triorder = 8
        tri = TriangularElement(triorder)

        def tri_equilateral_nodes_reference(self):
            # This is the old, more explicit, less general way of computing
            # the triangle nodes. Below, we compare its results with that of the
            # new routine.

            alpha_opt = [0.0000, 0.0000, 1.4152, 0.1001, 0.2751, 0.9800, 1.0999,
                    1.2832, 1.3648, 1.4773, 1.4959, 1.5743, 1.5770, 1.6223, 1.6258]
                      
            try:
                alpha = alpha_opt[self.order-1]
            except IndexError:
                alpha = 5/3

            from hedge.element import WarpFactorCalculator
            from math import sin, cos, pi

            warp = WarpFactorCalculator(self.order)

            edge1dir = numpy.array([1,0])
            edge2dir = numpy.array([cos(2*pi/3), sin(2*pi/3)])
            edge3dir = numpy.array([cos(4*pi/3), sin(4*pi/3)])

            for bary in self.equidistant_barycentric_nodes():
                lambda1, lambda2, lambda3 = bary

                # find equidistant (x,y) coordinates in equilateral triangle
                point = self.barycentric_to_equilateral(bary)

                # compute blend factors
                blend1 = 4*lambda1*lambda2 # nonzero on AB
                blend2 = 4*lambda3*lambda2 # nonzero on BC
                blend3 = 4*lambda3*lambda1 # nonzero on AC

                # calculate amount of warp for each node, for each edge
                warp1 = blend1*warp(lambda2 - lambda1)*(1 + (alpha*lambda3)**2)
                warp2 = blend2*warp(lambda3 - lambda2)*(1 + (alpha*lambda1)**2)
                warp3 = blend3*warp(lambda1 - lambda3)*(1 + (alpha*lambda2)**2)

                # return warped point
                yield point + warp1*edge1dir + warp2*edge2dir + warp3*edge3dir

        if False:
            outf = open("trinodes1.dat", "w")
            for ux in tri.equilateral_nodes():
                outf.write("%g\t%g\n" % tuple(ux))
            outf = open("trinodes2.dat", "w")
            for ux in tri_equilateral_nodes_reference(tri):
                outf.write("%g\t%g\n" % tuple(ux))

        for n1, n2 in zip(tri.equilateral_nodes(), 
                tri_equilateral_nodes_reference(tri)):
            self.assert_(la.norm(n1-n2) < 3e-15)

        def node_indices_2(order):
            for n in range(0, order+1):
                 for m in range(0, order+1-n):
                     yield m,n

        self.assert_(set(tri.node_tuples()) == set(node_indices_2(triorder)))
    # -------------------------------------------------------------------------
    def test_simp_basis_grad(self):
        """Do a simplistic FD-style check on the differentiation matrix"""
        from itertools import izip
        from hedge.element import \
                IntervalElement, \
                TriangularElement, \
                TetrahedralElement
        from random import uniform

        els = [
                (1, IntervalElement(5)), 
                (1, TriangularElement(8)), 
                (3,TetrahedralElement(7))]

        for err_factor, el in els:
            d = el.dimensions
            for i_bf, (bf, gradbf) in \
                    enumerate(izip(el.basis_functions(), el.grad_basis_functions())):
                for i in range(10):
                    base = -0.95
                    remaining = 1.90
                    r = numpy.zeros((d,))
                    for i in range(d):
                        rn = uniform(0, remaining)
                        r[i] = base+rn
                        remaining -= rn

                    from pytools import wandering_element
                    h = 1e-4
                    gradbf_v = numpy.array(gradbf(r))
                    approx_gradbf_v = numpy.array([
                        (bf(r+h*dir) - bf(r-h*dir))/(2*h)
                        for dir in [numpy.array(dir) for dir in wandering_element(d)]
                        ])
                    err = la.norm(approx_gradbf_v-gradbf_v, numpy.Inf)
                    #print el.dimensions, el.order, i_bf, err
                    self.assert_(err < err_factor*h)
    # -------------------------------------------------------------------------
    def test_tri_face_node_distribution(self):
        """Test whether the nodes on the faces of the triangle are distributed 
        according to the same proportions on each face.

        If this is not the case, then reusing the same face mass matrix
        for each face would be invalid.
        """

        from hedge.element import TriangularElement

        tri = TriangularElement(8)
        unodes = tri.unit_nodes()
        projected_face_points = []
        for face_i in tri.face_indices():
            start = unodes[face_i[0]]
            end = unodes[face_i[-1]]
            dir = end-start
            dir /= numpy.dot(dir, dir)
            pfp = numpy.array([numpy.dot(dir, unodes[i]-start) for i in face_i])
            projected_face_points.append(pfp)

        first_points =  projected_face_points[0]
        for points in projected_face_points[1:]:
            error = la.norm(points-first_points, numpy.Inf)
            self.assert_(error < 1e-15)
    # -------------------------------------------------------------------------
    def test_simp_face_normals_and_jacobians(self):
        """Check computed face normals and face jacobians on simplicial elements
        """
        from hedge.element import \
                IntervalElement, \
                TriangularElement, \
                TetrahedralElement
        from hedge.mesh import Triangle
        from hedge.tools import AffineMap
        from numpy import dot

        for el in [
                IntervalElement(3), 
                TetrahedralElement(1), 
                TriangularElement(4), 
                ]:
            for i in range(50):
                geo = el.geometry

                vertices = [numpy.random.randn(el.dimensions) 
                        for vi in range(el.dimensions+1)]
                #array = num.array
                #vertices = [array([-1, -1.0, -1.0]), array([1, -1.0, -1.0]), array([-1.0, 1, -1.0]), array([-1.0, -1.0, 1.0])]
                map = geo.get_map_unit_to_global(vertices)

                unodes = el.unit_nodes()
                nodes = [map(v) for v in unodes]

                all_vertex_indices = range(el.dimensions+1)

                for face_i, fvi, normal, jac in \
                        zip(el.face_indices(), 
                                geo.face_vertices(all_vertex_indices),
                                *geo.face_normals_and_jacobians(vertices, map)):
                    mapped_corners = [vertices[i] for i in fvi]
                    mapped_face_basis = [mc-mapped_corners[0] for mc in mapped_corners[1:]]

                    # face vertices must be among all face nodes
                    close_nodes = 0
                    for fi in face_i:
                        face_node = nodes[fi]
                        for mc in mapped_corners:
                            if la.norm(mc-face_node) < 1e-13:
                                close_nodes += 1

                    self.assert_(close_nodes == len(mapped_corners))

                    opp_node = (set(all_vertex_indices) - set(fvi)).__iter__().next()
                    mapped_opposite = vertices[opp_node]

                    if el.dimensions == 1:
                        true_jac = 1
                    elif el.dimensions == 2:
                        true_jac = la.norm(mapped_corners[1]-mapped_corners[0])/2
                    elif el.dimensions == 3:
                        from hedge.tools import orthonormalize
                        mapped_face_projection = numpy.array(
                                orthonormalize(mapped_face_basis))
                        projected_corners = (
                                [ numpy.zeros((2,))] 
                                + [dot(mapped_face_projection, v) for v in mapped_face_basis])
                        true_jac = abs(Triangle
                                .get_map_unit_to_global(projected_corners)
                                .jacobian)
                    else:
                        assert False, "this test does not support %d dimensions yet" % el.dimensions

                    #print abs(true_jac-jac)/true_jac
                    #print "aft, bef", la.norm(mapped_end-mapped_start),la.norm(end-start)

                    self.assert_(abs(true_jac - jac)/true_jac < 1e-13)
                    self.assert_(abs(la.norm(normal) - 1) < 1e-13)
                    for mfbv in mapped_face_basis:
                        self.assert_(abs(dot(normal, mfbv)) < 1e-13)

                    for mc in mapped_corners:
                        self.assert_(dot(mapped_opposite-mc, normal) < 0)
    # -------------------------------------------------------------------------
    def test_tri_map(self):
        """Verify that the mapping and node-building operations maintain triangle vertices"""
        from hedge.element import TriangularElement

        n = 8
        tri = TriangularElement(n)

        node_dict = dict((ituple, idx) for idx, ituple in enumerate(tri.node_tuples()))
        corner_indices = [node_dict[0,0], node_dict[n,0], node_dict[0,n]]
        unodes = tri.unit_nodes()
        corners = [unodes[i] for i in corner_indices]

        for i in range(10):
            vertices = [numpy.random.randn(2) for vi in range(3)]
            map = tri.geometry.get_map_unit_to_global(vertices)
            global_corners = [map(pt) for pt in corners]
            for gc, v in zip(global_corners, vertices):
                self.assert_(la.norm(gc-v) < 1e-12)
    # -------------------------------------------------------------------------
    def test_tri_map_jacobian_and_mass_matrix(self):
        """Verify whether tri map jacobians recover known values of triangle area"""
        from hedge.element import TriangularElement
        from math import sqrt, exp, pi

        for i in range(1,10):
            edata = TriangularElement(i)
            ones = numpy.ones((edata.node_count(),))
            unit_tri_area = 2
            error = la.norm(
                numpy.dot(ones,numpy.dot(edata.mass_matrix(), ones))-unit_tri_area)
            self.assert_(error < 1e-14)

        for i in range(10):
            vertices = [numpy.random.randn(2) for vi in range(3)]
            map = edata.geometry.get_map_unit_to_global(vertices)
            mat = numpy.zeros((2,2))
            mat[:,0] = (vertices[1] - vertices[0])
            mat[:,1] = (vertices[2] - vertices[0])
            tri_area = abs(la.det(mat)/2)
            tri_area_2 = abs(unit_tri_area*map.jacobian)
            self.assert_(abs(tri_area - tri_area_2)/tri_area < 1e-15)
    # -------------------------------------------------------------------------
    def no_test_tri_mass_mat_gauss(self):
        """Check the integral of a Gaussian on a disk using the mass matrix"""

        # This is a bad test, since it's never exact. The Gaussian has infinite support,
        # and this *does* matter numerically.

        from hedge.mesh import make_disk_mesh
        from hedge.element import TriangularElement
        from hedge.discretization import Discretization
        from math import sqrt, exp, pi

        sigma_squared = 1/219.3

        mesh = make_disk_mesh()
        discr = Discretization(make_disk_mesh(), TriangularElement(4), debug=True)
        f = discr.interpolate_volume_function(lambda x: exp(-x*x/(2*sigma_squared)))
        ones = discr.interpolate_volume_function(lambda x: 1)

        #discr.visualize_vtk("gaussian.vtk", [("f", f)])
        num_integral_1 = ones * (discr.mass_operator * f)
        num_integral_2 = f * (discr.mass_operator * ones)
        dim = 2
        true_integral = (2*pi)**(dim/2)*sqrt(sigma_squared)**dim
        err_1 = abs(num_integral_1-true_integral)
        err_2 = abs(num_integral_2-true_integral)
        self.assert_(err_1 < 1e-11)
        self.assert_(err_2 < 1e-11)
    # -------------------------------------------------------------------------
    def test_1d_mass_mat_trig(self):
        """Check the integral of some trig functions on an interval using the mass matrix"""
        from hedge.mesh import make_uniform_1d_mesh
        from hedge.element import IntervalElement
        from hedge.discretization import Discretization, integral
        from math import sqrt, pi, cos, sin
        from numpy import dot

        mesh = make_uniform_1d_mesh(-4*pi, 9*pi, 17, periodic=True)
        discr = Discretization(mesh, IntervalElement(8), debug=True)

        f = discr.interpolate_volume_function(lambda x: cos(x[0])**2)
        ones = discr.interpolate_volume_function(lambda x: 1)

        num_integral_1 = dot(ones, discr.mass_operator * f)
        num_integral_2 = dot(f, discr.mass_operator * ones)
        num_integral_3 = integral(discr, f)

        true_integral = 13*pi/2
        err_1 = abs(num_integral_1-true_integral)
        err_2 = abs(num_integral_2-true_integral)
        err_3 = abs(num_integral_3-true_integral)

        self.assert_(err_1 < 1e-10)
        self.assert_(err_2 < 1e-10)
        self.assert_(err_3 < 1e-10)
    # -------------------------------------------------------------------------
    def test_tri_mass_mat_trig(self):
        """Check the integral of some trig functions on a square using the mass matrix"""

        from hedge.mesh import make_square_mesh
        from hedge.element import TriangularElement
        from hedge.discretization import Discretization
        from math import sqrt, pi, cos, sin

        mesh = make_square_mesh(a=-pi, b=pi, max_area=(2*pi/10)**2/2)
        discr = Discretization(mesh, TriangularElement(8), debug=True)
        f = discr.interpolate_volume_function(lambda x: cos(x[0])**2*sin(x[1])**2)
        ones = discr.interpolate_volume_function(lambda x: 1)

        num_integral_1 = numpy.dot(ones, discr.mass_operator * f)
        num_integral_2 = numpy.dot(f, discr.mass_operator * ones)
        true_integral = pi**2
        err_1 = abs(num_integral_1-true_integral)
        err_2 = abs(num_integral_2-true_integral)
        #print err_1, err_2
        self.assert_(err_1 < 1e-10)
        self.assert_(err_2 < 1e-10)
    # -------------------------------------------------------------------------
    def test_tri_diff_mat(self):
        """Check differentiation matrix along the coordinate axes on a disk
        
        Uses sines as the function to differentiate.
        """
        from hedge.mesh import make_disk_mesh
        from hedge.element import TriangularElement
        from hedge.discretization import Discretization, cache_diff_results
        from math import sin, cos, sqrt


        for coord in [0, 1]:
            mesh = make_disk_mesh()
            discr = Discretization(make_disk_mesh(), TriangularElement(4), debug=True)
            f = discr.interpolate_volume_function(lambda x: sin(3*x[coord]))
            df = discr.interpolate_volume_function(lambda x: 3*cos(3*x[coord]))

            nabla = discr.nabla
            nabla[coord].do_not_warn()
            
            df_num = nabla[coord] * cache_diff_results(f)
            df_num_no_cache = nabla[coord] * f
            error = df_num - df
            #discr.visualize_vtk("diff-err.vtk",
                    #[("f", f), ("df", df), ("df_num", df_num), ("error", error)])

            linf_error = la.norm(df_num-df, numpy.Inf)
            #print linf_error
            self.assert_(linf_error < 3e-5)

            linf_error = la.norm(df_num_no_cache-df, numpy.Inf)
            self.assert_(linf_error < 3e-5)

            linf_error = la.norm(df_num_no_cache-df_num, numpy.Inf)
            self.assert_(linf_error < 2e-13)
    # -------------------------------------------------------------------------
    def test_2d_gauss_theorem(self):
        """Verify Gauss's theorem explicitly on a mesh"""

        from hedge.element import TriangularElement
        from hedge.tools import AffineMap
        from hedge.mesh import make_disk_mesh
        from hedge.flux import Flux
        from hedge.discretization import \
                Discretization, \
                pair_with_boundary, \
                cache_diff_results
        from math import sin, cos, sqrt, exp, pi
        from numpy import dot

        class NormalFlux(Flux):
            def __init__(self, coordinate):
                Flux.__init__(self)
                self.coordinate = coordinate

            def __call__(self, int_face, ext_face):
                return int_face.normal[self.coordinate]

        class ZeroFlux(Flux):
            def __call__(self, int_face, ext_face):
                return 0

        one_sided_x = (NormalFlux(0), ZeroFlux())
        one_sided_y = (NormalFlux(1), ZeroFlux())

        def f1(x):
            return sin(3*x[0])+cos(3*x[1])
        def f2(x):
            return sin(2*x[0])+cos(x[1])

        edata = TriangularElement(2)

        discr = Discretization(make_disk_mesh(), edata, debug=True)
        ones = discr.interpolate_volume_function(lambda x: 1)
        face_zeros = discr.boundary_zeros()

        f1_v = discr.interpolate_volume_function(f1)
        f2_v = discr.interpolate_volume_function(f2)

        f1_f = discr.interpolate_boundary_function(f1)
        f2_f = discr.interpolate_boundary_function(f2)

        dx_v = discr.nabla[0] * cache_diff_results(f1_v)
        dy_v = discr.nabla[1] * cache_diff_results(f2_v)

        int_div = \
                dot(ones, discr.mass_operator*dx_v) + \
                dot(ones, discr.mass_operator*dy_v)

        boundary_int = dot(
                discr.get_flux_operator(one_sided_x)*pair_with_boundary(f1_v, face_zeros) +
                discr.get_flux_operator(one_sided_y)*pair_with_boundary(f2_v, face_zeros),
                ones)

        #print abs(boundary_int-int_div)
        self.assert_(abs(boundary_int-int_div) < 5e-15)
    # -------------------------------------------------------------------------
    def test_simp_cubature(self):
        """Check that Grundmann-Moeller cubature works as advertised"""
        from pytools import generate_nonnegative_integer_tuples_summing_to_at_most
        from hedge.quadrature import SimplexCubature

        for dim in range(2,3+1):
            for s in range(3+1):
                cub = SimplexCubature(s, dim)
                for comb in generate_nonnegative_integer_tuples_summing_to_at_most(
                        2*s+1, dim):
                    f = Monomial(comb)
                    i_f = cub(f)
                    err = abs(i_f - f.theoretical_integral())
                    self.assert_(err < 2e-15)
    # -------------------------------------------------------------------------
    def test_simp_mass_and_diff_matrices_by_monomial(self):
        """Verify simplicial mass and differentiation matrices using monomials"""

        from hedge.element import \
                IntervalElement, \
                TriangularElement, \
                TetrahedralElement
        from pytools import generate_nonnegative_integer_tuples_summing_to_at_most

        from operator import add, mul

        thresh = 1e-13

        from numpy import dot
        for el in [
                IntervalElement(5),
                TriangularElement(3),
                TetrahedralElement(5),
                ]:
            for comb in generate_nonnegative_integer_tuples_summing_to_at_most(
                    el.order, el.dimensions):
                ones = numpy.ones((el.node_count(),))
                unodes = el.unit_nodes()
                f = Monomial(comb)
                f_n = numpy.array([f(x) for x in unodes])
                int_f_n = dot(ones, dot(el.mass_matrix(), f_n))
                int_f = f.theoretical_integral()
                err = la.norm(int_f - int_f_n)
                if err > thresh:
                    print "bad", el, comb, int_f, int_f_n, err
                self.assert_(err < thresh)

                dmats = el.differentiation_matrices()
                for i in range(el.dimensions):
                    df = f.diff(i)
                    df = numpy.array([df(x) for x in unodes])/2
                    df_n = dot(dmats[i], f_n)
                    err = la.norm(df - df_n, numpy.Inf)
                    if err > thresh:
                        print "bad-diff", comb, i, err
                    self.assert_(err < thresh)
    # -------------------------------------------------------------------------
    def test_simp_gauss_theorem(self):
        """Verify Gauss's theorem explicitly on simplicial elements"""

        from hedge.element import \
                IntervalElement, \
                TriangularElement, \
                TetrahedralElement
        from hedge.tools import AffineMap
        from operator import add
        from math import sin, cos, sqrt, exp, pi

        from numpy import dot

        def f1_1d(x):
            return sin(3*x[0])

        def f1_2d(x):
            return sin(3*x[0])+cos(3*x[1])
        def f2_2d(x):
            return sin(2*x[0])+cos(x[1])

        def f1_3d(x):
            return sin(3*x[0])+cos(3*x[1])+sin(1.9*x[2])
        def f2_3d(x):
            return sin(1.2*x[0])+cos(2.1*x[1])-cos(1.5*x[2])
        def f3_3d(x):
            return 5*sin(-0.2*x[0])-3*cos(x[1])+cos(x[2])

        #def f1_3d(x):
            #return 1
        #def f2_3d(x):
            #return 0
        #def f3_3d(x):
            #return 0

        def d(imap, coordinate, field):
            col = imap.matrix[:, coordinate]
            matrices = el.differentiation_matrices()
            return reduce(add, (coeff*dot(dmat, field)
                        for dmat, coeff in zip(matrices, col)))

        array = numpy.array

        intervals = [
                [array([-0.5]), array([17.])]
                ]

        triangles = [
                [array([-7.1687642250744492, 0.63058995062684642]), array([9.9744219044921199, 6.6530989283689781]), array([12.269380138171147, -17.529689194536481])],
                [array([-3.1285787297852634, -16.579403405465403]), array([-5.2882160938912515, -6.2209234150214137]), array([11.251223490342774, 4.6571427341871727])],
                [array([4.7407472917152553, -18.406868078408063]), array([1.8224524488556271, 11.551374404003361]), array([2.523148394963088, 1.632574414790982])],
                [array([-11.523714017493292, -14.2820557378961]), array([-0.44311816855771136, 19.572194735728861]), array([5.2855990566779445, -9.8743423935894388])],
                [array([1.113949150102217, -3.2255502625302639]), array([-13.028732972681315, 2.1525752429773379]), array([-2.3929000970202705, 6.2884649052982198])],
                [array([-8.0878061368549741, -14.604092423350167]), array([4.5339922477199996, 8.3770287646932022]), array([-5.2180549365480156, -1.9930760171433717])],
                [array([-1.9047012017294165, -3.6517882549544485]), array([3.1461902282192784, 5.7397849191668229]), array([-11.072761256907262, -8.3758508470287509])],
                [array([8.6609581113102934, 9.1121629958018566]), array([3.8230948675835497, -14.004679313330751]), array([10.975623610855521, 1.6267418698764553])],
                [array([13.959685276941629, -12.201892555481464]), array([-7.8057604576925499, -3.5283871457281757]), array([-0.41961743047735317, -3.2615635891671872])],
                [array([-9.8469907360335078, 6.0635407355366242]), array([7.8727080309703439, 7.634505157189091]), array([-2.7723038834027118, 8.5441656500931789])],
                ]
        tets = [
                #[make_random_vector(3, num.Float) for i in range(4)]
                #for j in range(10)
                [array([-0.087835976182812386, 8.4241880323369127, 2.6557808710807933]), array([5.4875560966799677, -7.5530368326122499, 8.4703868377747877]), array([-8.4888098806626751, 1.8369058453192324, -6.9041468708803713]), array([17.327527449499168, -9.0586108433594319, 5.7459746913914636])],
                [array([16.993689961344593, -12.116305360441197, -12.711045554409088]), array([-2.0324332275643817, -5.0524187595904335, 5.9257028535230383]), array([6.4221144729287687, -7.2496949199427245, -1.1590391996379827]), array([-5.7529432953399171, -6.9587987820990262, 3.7223773892240426])],
                [array([-0.4423263927732628, -1.6306971591009138, -1.2554069824001064]), array([-9.1171749892163785, 14.232868970928301, 4.6548620163014505]), array([16.554360867165595, -2.1451702825571202, -1.9050837421951314]), array([-8.7455417971698139, 19.016251630886945, -15.137691422305545])],
                [array([-1.9251811954429843, -4.5369007736338665, 9.2675942450331963]), array([-13.586778017089083, -3.6666239130220553, -14.095112617514117]), array([-15.014799506040006, -3.4363888726140681, -0.85237426213437206]), array([6.3854978041452597, 13.293981904633554, -7.8432774486183146])],
                [array([-6.761839340374304, 14.864784704377955, 1.574274771089831]), array([-0.1823468063801317, -21.892423945260102, 11.565172070570537]), array([-0.14658389181168049, 13.07241603902848, 7.2652184007323042]), array([-20.35574011769496, 14.816503793175773, -7.2800214849607254])],
                [array([23.294362873156878, 13.644282203469114, 10.383738204469243]), array([-19.792088993452555, 0.4209925297886693, -7.3933945447515388]), array([-2.832898385995708, -1.6480401382241885, -6.2689214950820924]), array([-0.081772347748623617, -3.3803599922239673, -19.614368674546114])],
                [array([0.43913744703796659, -16.473036116412242, -0.8653853759721295]), array([-7.3270752283484297, -0.97723025169973787, 2.1330514627504464]), array([3.8730334021748307, -9.0983850278388143, 3.3578300089831501]), array([18.639504439820936, 20.594835769217696, -10.666261239487298])],
                [array([-12.786230591302058, -9.2931510923111169, -2.1598642469378935]), array([-4.0458439207057459, -9.0298998997705144, -0.11666215074316685]), array([7.5023999981398424, 4.8603369473110583, 2.1813627427875013]), array([2.9579841500551272, -22.563123335973565, 10.335559822513606])],
                [array([-7.7732699602949893, 15.816977096296963, -6.8826683632918728]), array([7.6233333630240256, -9.3309869383569026, 0.50189282953625991]), array([-11.272342858699034, 1.089016041114454, -6.0359393299451476]), array([-6.4746449930954348, -0.026130504314747997, -2.2786267101817677])],
                [array([-18.243993907118757, 5.0646875774948974, -9.2110046334596856]), array([-8.1550804560957264, -3.1021806460634913, 7.5622831439916105]), array([19.460768761970783, 17.494565076685859, 16.295621155355697]), array([4.6186236213250131, -1.3869183721072562, -0.2159066724152843])],
                ]

        for el_geoms, el, f in [
                (intervals, IntervalElement(9), [f1_1d]),
                (triangles, TriangularElement(9), [f1_2d, f2_2d]),
                (tets, TetrahedralElement(1), [f1_3d, f2_3d, f3_3d]),
                ]:
            for vertices in el_geoms:
                ones = numpy.ones((el.node_count(),))
                face_ones = numpy.ones((len(el.face_indices()[0]),))

                map = el.geometry.get_map_unit_to_global(vertices)
                imap = map.inverted()

                mapped_points = [map(node) for node in el.unit_nodes()]

                f_n = [numpy.array([fi(x) for x in mapped_points])
                        for fi in f]
                df_n = [d(imap, i, f_n[i]) for i, fi_n in enumerate(f_n)]

                int_div_f = abs(map.jacobian)*sum(
                        dot(ones, dot(el.mass_matrix(), dfi_n)) for dfi_n in df_n)

                if False:
                    boundary_comp = [
                            array([
                                fjac 
                                * dot(face_ones,
                                    dot(el.face_mass_matrix(),
                                        num.take(fi_n, face_indices)) )
                                * n_coord
                                for fi_n, n_coord in zip(f_n, n)])
                            for face_indices, n, fjac
                            in zip(el.face_indices(), *el.face_normals_and_jacobians(vertices, map))
                            ]

                boundary_sum = sum(
                        sum(
                            fjac 
                            * dot(face_ones,
                                dot(el.face_mass_matrix(), 
                                    numpy.take(fi_n, face_indices)))
                            * n_coord
                            for fi_n, n_coord in zip(f_n, n))
                        for face_indices, n, fjac
                        in zip(el.face_indices(), 
                            *el.geometry.face_normals_and_jacobians(vertices, map))
                        )

                #print el.face_normals_and_jacobians(map)[1]
                #print 'mp', [mapped_points[fi] for fi in el.face_indices()[2]]
                #print num.take(f_n[0], el.face_indices()[2])
                #print 'bc', boundary_comp
                #print 'bs', boundary_sum
                #print 'idiv', int_div_f
                #print abs(boundary_sum-int_div_f)
                self.assert_(abs(boundary_sum-int_div_f) < 1e-12)
    # -------------------------------------------------------------------------
    def test_simp_orthogonality(self):
        """Test orthogonality of simplicial bases using Grundmann-Moeller cubature"""
        from hedge.quadrature import SimplexCubature
        from hedge.element import TriangularElement, TetrahedralElement

        from numpy import dot

        for order, ebound in [
                (1, 2e-15),
                (2, 5e-15),
                (3, 1e-14),
                #(4, 3e-14),
                #(7, 3e-14),
                #(9, 2e-13),
                ]:
            for ldis in [TriangularElement(order), TetrahedralElement(order)]:
                cub = SimplexCubature(order, ldis.dimensions)
                basis = ldis.basis_functions()

                maxerr = 0
                for i, f in enumerate(basis):
                    for j, g in enumerate(basis):
                        if i == j:
                            true_result = 1
                        else:
                            true_result = 0
                        result = cub(lambda x: f(x)*g(x))
                        err = abs(result-true_result)
                        maxerr = max(maxerr, err)
                        if err > ebound:
                            print "bad", order,i,j, err
                        self.assert_(err < ebound)
                #print order, maxerr
    # -------------------------------------------------------------------------
    def test_1d_mass_matrix_vs_quadrature(self):
        """Check that a 1D mass matrix for Legendre-Gauss points gives the right weights"""
        from hedge.quadrature import LegendreGaussQuadrature
        from hedge.polynomial import legendre_vandermonde

        for n in range(13):
            lgq = LegendreGaussQuadrature(n)
            vdm = legendre_vandermonde(lgq.points, n)
            mass_mat = la.inv(numpy.dot(vdm, vdm.T))
            ones = numpy.ones((mass_mat.shape[0],))
            self.assert_(la.norm(
                la.solve(numpy.dot(vdm, vdm.T), ones) -
                    numpy.array(lgq.weights), numpy.Inf) < 2e-14)
    # -------------------------------------------------------------------------
    def test_mapping_differences_tri(self):
        """Check that triangle interpolation is independent of mapping to reference
        """
        from hedge.element import TriangularElement
        from random import random
        from pytools import generate_permutations

        def shift(list):
            return list[1:] + [list[0]]

        class LinearCombinationOfFunctions:
            def __init__(self, coefficients, functions, premap):
                self.coefficients = coefficients
                self.functions = functions
                self.premap = premap

            def __call__(self, x):
                return sum(coeff*f(self.premap(x)) for coeff, f in 
                        zip(self.coefficients, self.functions))

        def random_barycentric_coordinates(dim):
            remain = 1
            coords = []
            for i in range(dim):
                coords.append(random() * remain)
                remain -= coords[-1]
            coords.append(remain)
            return coords

        tri = TriangularElement(5)

        for trial_number in range(10):
            vertices = [numpy.random.randn(2) for vi in range(3)]
            map = tri.geometry.get_map_unit_to_global(vertices)
            nodes = [map(node) for node in tri.unit_nodes()]
            node_values = numpy.array([random() for node in nodes])

            functions = []
            for pvertices in generate_permutations(vertices):
                pmap = tri.geometry.get_map_unit_to_global(pvertices)
                pnodes = [pmap(node) for node in tri.unit_nodes()]

                # map from pnode# to node#
                nodematch = {}
                for pi, pn in enumerate(pnodes):
                    for i, n in enumerate(nodes):
                        if la.norm(n - pn) < 1e-13:
                            nodematch[pi] = i
                            break

                pnode_values = numpy.array([node_values[nodematch[pi]] 
                        for pi in range(len(nodes))])

                interp_f = LinearCombinationOfFunctions(
                        la.solve(tri.vandermonde(), pnode_values),
                        tri.basis_functions(),
                        pmap.inverted())

                # verify interpolation property
                #for n, nv in zip(pnodes, pnode_values):
                    #self.assert_(abs(interp_f(n) - nv) < 1e-13)

                functions.append(interp_f)

            for subtrial_number in range(15):
                pt_in_element = sum(
                        coeff*vertex
                        for coeff, vertex in zip(
                            random_barycentric_coordinates(2),
                            vertices))
                f_values = [f(pt_in_element) for f in functions]
                avg = sum(f_values) / len(f_values)
                err = [abs(fv-avg) for fv in f_values]
                self.assert_(max(err) < 5e-13)
    # -------------------------------------------------------------------------
    def test_interior_fluxes_tri(self):
        """Check triangle surface integrals computed using interior fluxes
        against their known values.
        """

        from math import pi, sin, cos

        def round_trip_connect(start, end):
            for i in range(start, end):
                yield i, i+1
            yield end, start

        a = -pi
        b = pi
        points = [
                (a,0), (b,0), 
                (a,-1), (b,-1),
                (a,1), (b,1)
                ]
                
        import meshpy.triangle as triangle

        mesh_info = triangle.MeshInfo()
        mesh_info.set_points(points)
        mesh_info.set_facets(
                [(0,1),(1,3),(3,2),(2,0),(0,4),(4,5),(1,5)]
                )

        mesh_info.regions.resize(2)
        mesh_info.regions[0] = [
                0,-0.5, # coordinate
                1, # lower element tag
                0.1, # max area
                ]
        mesh_info.regions[1] = [
                0,0.5, # coordinate
                2, # upper element tag
                0.01, # max area
                ]

        generated_mesh = triangle.build(mesh_info, 
                attributes=True,
                volume_constraints=True)
        #triangle.write_gnuplot_mesh("mesh.dat", generated_mesh)

        def element_tagger(el):
            if generated_mesh.element_attributes[el.id] == 1:
                return ["upper"]
            else:
                return ["lower"]

        from hedge.mesh import make_conformal_mesh
        mesh = make_conformal_mesh(
                generated_mesh.points,
                generated_mesh.elements,
                element_tagger=element_tagger)

        from hedge.element import TriangularElement
        from hedge.discretization import Discretization, ones_on_volume
        discr = Discretization(mesh, TriangularElement(4), debug=True)

        u_i = numpy.multiply(
                discr.interpolate_volume_function(lambda x: sin(x[0]-x[1])),
                ones_on_volume(discr, "lower"))
        u_o = numpy.multiply(
                discr.interpolate_volume_function(lambda x: cos(x[0]-x[1])),
                ones_on_volume(discr, "upper"))
        u = u_i + u_o

        #discr.visualize_vtk("dual.vtk", [("u", u)])

        from hedge.flux import make_normal, FluxScalarPlaceholder
        fluxu = FluxScalarPlaceholder()
        res = discr.get_flux_operator(
                (fluxu.int - fluxu.ext)*make_normal(discr.dimensions)[1]) * u

        ones = discr.interpolate_volume_function(lambda x: 1)
        self.assert_(abs(numpy.dot(res, ones)) < 5e-14)
    # -------------------------------------------------------------------------
    def test_interior_fluxes_tet(self):
        """Check tetrahedron surface integrals computed using interior fluxes
        against their known values.
        """

        import meshpy.tet as tet
        from math import pi, sin, cos

        mesh_info = tet.MeshInfo()

        # construct a two-box extrusion of this base
        base = [(-pi,-pi,0), (pi,-pi,0), (pi,pi,0), (-pi,pi,0)]

        # first, the nodes
        mesh_info.set_points(
                base
                +[(x,y,z+pi) for x,y,z in base]
                +[(x,y,z+pi+1) for x,y,z in base]
                )

        # next, the facets

        # vertex indices for a box missing the -z face
        box_without_minus_z = [ 
            [4,5,6,7],
            [0,4,5,1],
            [1,5,6,2],
            [2,6,7,3],
            [3,7,4,0],
            ]

        def add_to_all_vertex_indices(facets, increment):
            return [[pt+increment for pt in facet] for facet in facets]

        mesh_info.set_facets(
            [[0,1,2,3]] # base
            +box_without_minus_z # first box
            +add_to_all_vertex_indices(box_without_minus_z, 4) # second box
            )

        # set the volume properties -- this is where the tet size constraints are
        mesh_info.regions.resize(2)
        mesh_info.regions[0] = [0,0,pi/2, # point in volume -> first box
                0, # region tag (user-defined number)
                0.5, # max tet volume in region
                ]
        mesh_info.regions[1] = [0,0,pi+0.5, # point in volume -> second box
                1, # region tag (user-defined number, arbitrary)
                0.1, # max tet volume in region
                ]

        generated_mesh = tet.build(mesh_info, attributes=True, volume_constraints=True)
        #mesh.write_vtk("sandwich-mesh.vtk")


        def element_tagger(el):
            if generated_mesh.element_attributes[el.id] == 1:
                return ["upper"]
            else:
                return ["lower"]

        from hedge.mesh import make_conformal_mesh
        mesh = make_conformal_mesh(
                generated_mesh.points,
                generated_mesh.elements,
                element_tagger=element_tagger)

        from hedge.element import TetrahedralElement
        from hedge.discretization import Discretization, ones_on_volume
        discr = Discretization(mesh, TetrahedralElement(4), debug=True)

        u_l = numpy.multiply(
                discr.interpolate_volume_function(lambda x: sin(x[0]-x[1]+x[2])),
                ones_on_volume(discr, "lower"))
        u_u = numpy.multiply(
                discr.interpolate_volume_function(lambda x: cos(x[0]-x[1]+x[2])),
                ones_on_volume(discr, "upper"))
        u = u_l + u_u

        # visualize the produced field
        #from hedge.visualization import SiloVisualizer
        #vis = SiloVisualizer(discr)
        #vis("sandwich.silo", [("u_l", u_l), ("u_u", u_u)], expressions=[("u", "u_l+u_u")],
                #write_coarse_mesh=True)

        # make sure the surface integral of the difference 
        # between top and bottom is zero
        from hedge.flux import make_normal, FluxScalarPlaceholder
        fluxu = FluxScalarPlaceholder()
        res = discr.get_flux_operator(
                (fluxu.int - fluxu.ext)*make_normal(discr.dimensions)[1]) * u
        ones = discr.interpolate_volume_function(lambda x: 1)
        self.assert_(abs(numpy.dot(res,ones)) < 5e-14)
    # -------------------------------------------------------------------------
    def test_symmetry_preservation_2d(self):
        """Test whether we preserve symmetry in a symmetric 2D advection problem"""
        from numpy import dot

        def make_mesh():
            array = numpy.array

            #
            #    1---8---2
            #    |7 /|\ 1|
            #    | / | \ |
            #    |/ 6|0 \|
            #    5---4---7
            #    |\ 5|3 /|
            #    | \ | / |
            #    |4 \|/ 2|
            #    0---6---3
            #
            points = [
                    array([-0.5, -0.5]), 
                    array([-0.5, 0.5]), 
                    array([0.5, 0.5]), 
                    array([0.5, -0.5]), 
                    array([0.0, 0.0]), 
                    array([-0.5, 0.0]), 
                    array([0.0, -0.5]), 
                    array([0.5, 0.0]), 
                    array([0.0, 0.5])]

            elements = [
                    [8,7,4],
                    [8,7,2],
                    [6,7,3],
                    [7,4,6],
                    [5,6,0],
                    [5,6,4],
                    [5,8,4],
                    [1,5,8],
                    ]

            def boundary_tagger(vertices, el, face_nr):
                if dot(el.face_normals[face_nr], v) < 0:
                    return ["inflow"]
                else:
                    return ["outflow"]

            from hedge.mesh import make_conformal_mesh
            return make_conformal_mesh(points, elements, boundary_tagger)

        from hedge.discretization import Discretization, SymmetryMap
        from hedge.element import TriangularElement
        from hedge.timestep import RK4TimeStepper
        from math import sqrt
        from hedge.operators import StrongAdvectionOperator
        from hedge.data import TimeDependentGivenFunction

        v = numpy.array([-1,0])

        mesh = make_mesh()
        discr = Discretization(mesh, TriangularElement(4), debug=True)

        def f(x):
            if x < 0.5: return 0
            else: return (x-0.5)

        def u_analytic(x, t):
            return f(-dot(v, x)+t)

        u = discr.interpolate_volume_function(lambda x: u_analytic(x, 0))

        sym_map = SymmetryMap(discr, 
                lambda x: numpy.array([x[0], -x[1]]),
                {0:3, 2:1, 5:6, 7:4})

        for flux_type in StrongAdvectionOperator.flux_types:
            stepper = RK4TimeStepper()
            op = StrongAdvectionOperator(discr, v, 
                    inflow_u=TimeDependentGivenFunction(u_analytic),
                    flux_type=flux_type)

            dt = discr.dt_factor(op.max_eigenvalue())
            nsteps = int(1/dt)
            for step in xrange(nsteps):
                u = stepper(u, step*dt, dt, op.rhs)
                sym_error_u = u-sym_map(u)
                sym_error_u_l2 = sqrt(dot(sym_error_u, (discr.mass_operator*sym_error_u)))
                self.assert_(sym_error_u_l2 < 4e-15)
    # -------------------------------------------------------------------------
    def test_convergence_advec_2d(self):
        """Test whether 2D advection actually converges"""

        from hedge.mesh import make_disk_mesh, make_regular_rect_mesh
        from hedge.discretization import Discretization, pair_with_boundary
        from hedge.element import TriangularElement
        from hedge.timestep import RK4TimeStepper
        from hedge.tools import EOCRecorder
        from math import sin, pi, sqrt
        from hedge.operators import StrongAdvectionOperator
        from hedge.data import TimeDependentGivenFunction

        v = numpy.array([0.27,0])
        norm_a = la.norm(v)

        from numpy import dot

        def f(x):
            return sin(x)

        def u_analytic(x, t):
            return f((-dot(v, x)/norm_a+t*norm_a))

        def boundary_tagger(vertices, el, face_nr):
            if dot(el.face_normals[face_nr], v) < 0:
                return ["inflow"]
            else:
                return ["outflow"]

        for mesh in [
                # non-periodic
                make_disk_mesh(r=pi, boundary_tagger=boundary_tagger, max_area=0.5),
                # periodic
                make_regular_rect_mesh(a=(0,0), b=(2*pi, 1), n=(8,4),
                    periodicity=(True, False),
                    boundary_tagger=boundary_tagger, 
                    )
                ]:
            for flux_type in StrongAdvectionOperator.flux_types:
                for direct_flux in [True, False]:
                    eoc_rec = EOCRecorder()

                    for order in [1,2,3,4,5,6]:
                        discr = Discretization(mesh, TriangularElement(order), debug=True)
                        op = StrongAdvectionOperator(discr, v, 
                                inflow_u=TimeDependentGivenFunction(u_analytic),
                                flux_type=flux_type, direct_flux=direct_flux)

                        u = discr.interpolate_volume_function(lambda x: u_analytic(x, 0))
                        dt = discr.dt_factor(norm_a)
                        nsteps = int(1/dt)

                        stepper = RK4TimeStepper()
                        for step in range(nsteps):
                            u = stepper(u, step*dt, dt, op.rhs)

                        u_true = discr.interpolate_volume_function(
                                lambda x: u_analytic(x, nsteps*dt))
                        error = u-u_true
                        error_l2 = sqrt(dot(error, discr.mass_operator*error))
                        eoc_rec.add_data_point(order, error_l2)

                    if False:
                        print "%s\n%s\n" % (flux_type.upper(), "-" * len(flux_type))
                        print eoc_rec.pretty_print(abscissa_label="Poly. Order", 
                                error_label="L2 Error")

                    self.assert_(eoc_rec.estimate_order_of_convergence()[0,1] > 4)
                    self.assert_(eoc_rec.estimate_order_of_convergence(2)[-1,1] > 10)
    # -------------------------------------------------------------------------
    def test_all_periodic_no_boundary(self):
        """Test that an all-periodic brick has no boundary."""
        from hedge.mesh import make_box_mesh, TAG_ALL

        mesh = make_box_mesh(periodicity=(True,True,True))

        def count(iterable):
            result = 0
            for i in iterable:
                result += 1
            return result

        self.assert_(count(mesh.tag_to_boundary[TAG_ALL]) == 0)
    # -------------------------------------------------------------------------
    def test_elliptic(self):
        """Test various properties of elliptic operators."""

        from hedge.tools import unit_vector
        def matrix_rep(op):
            h,w = op.shape
            mat = numpy.zeros(op.shape)
            for j in range(w):
                mat[:,j] = op(unit_vector(w, j))
            return mat

        def check_grad_mat():
            grad_mat = op.grad_matrix()

            #print len(discr), grad_mat.nnz, type(grad_mat)
            for i in range(10):
                u = numpy.random.randn(len(discr))

                mat_result = grad_mat * u
                op_result = numpy.hstack(op.grad(u))

                err = la.norm(mat_result-op_result)*la.norm(op_result)
                self.assert_(
                        la.norm(mat_result-op_result)*la.norm(op_result)
                        < 1e-5)

        def check_matrix_tgt():
            big = num.zeros((20, 20), flavor=num.SparseBuildMatrix)
            small = num.array([[1,2,3],[4,5,6],[7,8,9]])
            print small
            from hedge._internal import MatrixTarget
            tgt = MatrixTarget(big, 4, 4)
            tgt.begin(small.shape[0], small.shape[1])
            print "YO"
            tgt.add_coefficients(4, 4, small)
            print "DUDE"
            tgt.finalize()
            print big

        import pymbolic
        v_x = pymbolic.var("x")
        truesol = pymbolic.parse("math.sin(x[0]**2*x[1]**2)")
        truesol_c = pymbolic.compile(truesol, variables=["x"])
        rhs = pymbolic.simplify(pymbolic.laplace(truesol, [v_x[0], v_x[1]]))
        rhs_c = pymbolic.compile(rhs, variables=["x"])

        from hedge.mesh import make_disk_mesh, TAG_ALL, TAG_NONE
        mesh = make_disk_mesh(r=0.5, max_area=0.1, faces=20)
        mesh = mesh.reordered_by("cuthill")

        try:
            import pyublasext
            can_solve = True
        except ImportError:
            can_solve = False

        from hedge.tools import EOCRecorder
        eocrec = EOCRecorder()
        for order in [1,2,3,4,5]:
            for flux in ["ldg", "ip"]:
                from hedge.discretization import Discretization, norm
                from hedge.element import TriangularElement
                discr = Discretization(mesh, TriangularElement(order), debug=True)

                from hedge.data import GivenFunction
                from hedge.operators import WeakPoissonOperator
                op = WeakPoissonOperator(discr, 
                        dirichlet_tag=TAG_ALL,
                        dirichlet_bc=GivenFunction(truesol_c),
                        neumann_tag=TAG_NONE, 
                        flux=flux)

                if order <= 3:
                    mat = matrix_rep(op)
                    sym_err = la.norm(mat-mat.T)
                    self.assert_(sym_err<1e-12)
                    check_grad_mat()

                if can_solve:
                    truesol_v = discr.interpolate_volume_function(truesol_c)
                    a_inv = pyublasext.CGOperator.make(-op, 40000, 1e-10)
                    sol_v = -a_inv(op.prepare_rhs(GivenFunction(rhs_c)))

                    eocrec.add_data_point(order, norm(discr, sol_v-truesol_v))

        if can_solve:
            #print eocrec.pretty_print()
            self.assert_(eocrec.estimate_order_of_convergence()[0,1] > 8)
    # -------------------------------------------------------------------------
    def test_projection(self):
        """Test whether projection between different orders works"""

        from hedge.mesh import make_disk_mesh
        from hedge.discretization import Discretization, Projector, norm
        from hedge.element import TriangularElement
        from hedge.tools import EOCRecorder
        from math import sin, pi, sqrt

        from numpy import dot

        a = numpy.array([1,3])

        def u_analytic(x):
            return sin(dot(a, x))

        mesh = make_disk_mesh(r=pi, max_area=0.5)

        discr2 = Discretization(mesh, TriangularElement(2), debug=True)
        discr5 = Discretization(mesh, TriangularElement(5), debug=True)
        p2to5 = Projector(discr2, discr5)
        p5to2 = Projector(discr5, discr2)

        u2 = discr2.interpolate_volume_function(u_analytic)
        u2_i = p5to2(p2to5(u2))
        self.assert_(norm(discr2, u2-u2_i) < 3e-15)
    # -------------------------------------------------------------------------
    def test_filter(self):
        """Exercise mode-based filtering."""

        from hedge.mesh import make_disk_mesh
        from hedge.discretization import Discretization, integral, norm
        from hedge.element import TriangularElement
        from math import sin

        mesh = make_disk_mesh(r=3.4, max_area=0.5)
        discr = Discretization(mesh, TriangularElement(5), debug=True)

        from hedge.discretization import Filter, ExponentialFilterResponseFunction
        half_filter = Filter(discr, lambda mid, ldis: 0.5)
        for fmat in half_filter.filter_matrices:
            n,m = fmat.shape
            self.assert_(la.norm(fmat - 0.5*numpy.eye(n, m)) < 2e-15)

        from numpy import dot

        def test_freq(n):
            a = numpy.array([1,n])

            def u_analytic(x):
                return sin(dot(a, x))

            exp_filter = Filter(discr, ExponentialFilterResponseFunction(0.9, 3))

            u = discr.interpolate_volume_function(u_analytic)
            filt_u = exp_filter(u)

            int_error = abs(integral(discr, u) - integral(discr, filt_u))
            l2_ratio = norm(discr, filt_u) / norm(discr, u)
            self.assert_(int_error < 5e-15)
            self.assert_(0.975 < l2_ratio < 0.99999)

        test_freq(3)
        test_freq(5)
        test_freq(9)
        test_freq(17)




if __name__ == '__main__':
    unittest.main()



"""Visualization for global DG functions. Supports VTK, Silo, etc."""

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import hedge.tools 
import numpy




class Visualizer(object):
    pass




# legacy vtk ------------------------------------------------------------------
def _three_vector(x):
    if len(x) == 3:
        return x
    elif len(x) == 2:
        return x[0], x[1], 0.
    elif len(x) == 1:
        return x[0], 0, 0.




class LegacyVtkFile(object):
    def __init__(self, pathname, structure, description="Hedge visualization"):
        self.pathname = pathname
        self.structure = structure
        self.description = description

        self.pointdata = []
        self.is_closed = False

    def __fin__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        if not self.is_closed:
            from pyvtk import PointData, VtkData
            vtk = VtkData(self.structure, 
                    self.description, 
                    PointData(*self.pointdata))
            vtk.tofile(self.pathname)
            self.is_closed = True
            



class LegacyVtkVisualizer(Visualizer):
    def __init__(self, discr):
        from pyvtk import PolyData

        points = [_three_vector(p) for p in discr.nodes]
        polygons = []

        for eg in discr.element_groups:
            ldis = eg.local_discretization
            for el, (el_start, el_stop) in zip(eg.members, eg.ranges):
                polygons += [[el_start+j for j in element] 
                        for element in ldis.get_submesh_indices()]

        self.structure = PolyData(points=points, polygons=polygons)

    def make_file(self, pathname, pcontext=None):
        if pcontext is not None:
            if len(pcontext.ranks) > 1:
                raise RuntimeError, "Legacy VTK does not suport parallel visualization"
        return LegacyVtkFile(pathname+".vtk", self.structure)

    def add_data(self, vtkfile, scalars=[], vectors=[], scale_factor=1):
        from pyvtk import Scalars, Vectors
        import numpy

        vtkfile.pointdata.extend(
                Scalars(numpy.array(scale_factor*field), 
                    name=name, lookup_table="default") 
                for name, field in scalars)
        vtkfile.pointdata.extend(
                Vectors([_three_vector(scale_factor*v) 
                    for v in zip(field)], name=name)
                for name, field in vectors)




# xml vtk ---------------------------------------------------------------------
class VtkFile(hedge.tools.Closable):
    def __init__(self, pathname, grid, filenames=None, compressor=None):
        hedge.tools.Closable.__init__(self)
        self.pathname = pathname
        self.grid = grid
        self.compressor = compressor

    def get_head_pathname(self):
        return self.pathname

    def do_close(self):
        from pytools import assert_not_a_file
        assert_not_a_file(self.pathname)

        from hedge.vtk import InlineXMLGenerator, AppendedDataXMLGenerator

        outf = file(self.pathname, "w")
        AppendedDataXMLGenerator(self.compressor)(self.grid).write(outf)
        #InlineXMLGenerator(self.compressor)(self.grid).write(outf)
        outf.close()





class ParallelVtkFile(VtkFile):
    def __init__(self, pathname, grid, index_pathname, pathnames=None, compressor=None):
        VtkFile.__init__(self, pathname, grid)
        self.index_pathname = index_pathname
        self.pathnames = pathnames
        self.compressor = compressor

    def get_head_pathname(self):
        return self.index_pathname

    def do_close(self):
        VtkFile.do_close(self)

        from hedge.vtk import ParallelXMLGenerator

        outf = file(self.index_pathname, "w")
        ParallelXMLGenerator(self.pathnames)(self.grid).write(outf)
        outf.close()






class VtkVisualizer(Visualizer, hedge.tools.Closable):
    def __init__(self, discr, pcontext=None, basename=None, compressor=None):
        hedge.tools.Closable.__init__(self)

        from pytools import assert_not_a_file

        if basename is not None:
            self.pvd_name = basename+".pvd"
            assert_not_a_file(self.pvd_name)
        else:
            self.pvd_name = None

        self.pcontext = pcontext
        self.compressor = compressor

        if self.pcontext is None or self.pcontext.is_head_rank:
            self.timestep_to_pathnames = {}
        else:
            self.timestep_to_pathnames = None

        from hedge.vtk import UnstructuredGrid, DataArray, \
                VTK_LINE, VTK_TRIANGLE, VTK_TETRA, VF_LIST_OF_VECTORS
        from hedge.mesh import Interval, Triangle, Tetrahedron

        # For now, we use IntVector here because the Python allocator
        # is somewhat reluctant to return allocated chunks of memory
        # to the OS.
        from hedge._internal import IntVector
        cells = IntVector()
        cell_types = IntVector()

        for eg in discr.element_groups:
            ldis = eg.local_discretization
            smi = ldis.get_submesh_indices()

            cells.reserve(len(cells)+len(smi)*len(eg.members))
            for el, (el_start, el_stop) in zip(eg.members, eg.ranges):
                for element in smi:
                    for j in element:
                        cells.append(el_start+j)

            if ldis.geometry is Interval:
                vtk_eltype = VTK_LINE
            elif ldis.geometry is Triangle:
                vtk_eltype = VTK_TRIANGLE
            elif ldis.geometry is Tetrahedron:
                vtk_eltype = VTK_TETRA
            else:
                raise RuntimeError, "unsupported element type: %s" % ldis.geometry

            cell_types.extend([vtk_eltype] * len(smi) * len(eg.members))

        self.grid = UnstructuredGrid(
                (len(discr), 
                    DataArray("points", discr.nodes, vector_format=VF_LIST_OF_VECTORS)),
                cells, cell_types)


    def update_pvd(self):
        if self.pvd_name and self.timestep_to_pathnames:
            from hedge.vtk import XMLRoot, XMLElement, make_vtkfile

            collection = XMLElement("Collection")

            vtkf = make_vtkfile(collection.tag, compressor=None)
            xmlroot = XMLRoot(vtkf)

            vtkf.add_child(collection)

            tsteps = self.timestep_to_pathnames.keys()
            tsteps.sort()
            for i, time in enumerate(tsteps):
                for part, pathname in enumerate(self.timestep_to_pathnames[time]):
                    collection.add_child(XMLElement(
                        "DataSet",
                        timestep=time, part=part, file=pathname))
            outf = open(self.pvd_name, "w")
            xmlroot.write(outf)
            outf.close()

    def do_close(self):
        self.update_pvd()

    def make_file(self, pathname):
        """FIXME

        An appropriate extension (including the dot) is automatically
        appended to `pathname'.
        """
        if self.pcontext is None or len(self.pcontext.ranks) == 1:
            return VtkFile(pathname+"."+self.grid.vtk_extension(), 
                    self.grid.copy(),
                    compressor=self.compressor
                    )
        else:
            filename_pattern = (
                    pathname + "-%05d." + self.grid.vtk_extension())
            if self.pcontext.is_head_rank:
                return ParallelVtkFile(
                        filename_pattern % self.pcontext.rank,
                        self.grid.copy(), 
                        index_pathname="%s.p%s" % (
                            pathname, self.grid.vtk_extension()),
                        pathnames=[
                            filename_pattern % rank for rank in self.pcontext.ranks],
                       compressor=self.compressor 
                       )
            else:
                return VtkFile(
                        filename_pattern % self.pcontext.rank, 
                        self.grid.copy(),
                        compressor=self.compressor
                        )

    def register_pathname(self, time, pathname):
        if time is not None and self.timestep_to_pathnames is not None:
            self.timestep_to_pathnames.setdefault(time, []).append(pathname)

            # When we are run under MPI and cancelled by Ctrl+C, destructors
            # do not get called. Therefore, we just spend the (hopefully negligible)
            # time to update the PVD index every few data additions.
            if len(self.timestep_to_pathnames) % 5 == 0:
                self.update_pvd()

    def add_data(self, visf, variables=[], scalars=[], vectors=[], time=None, step=None,
            scale_factor=1):
        if scalars or vectors:
            import warnings
            warnings.warn("`scalars' and `vectors' arguments are deprecated",
                    DeprecationWarning)
            variables = scalars + vectors

        from hedge.vtk import DataArray, VF_LIST_OF_COMPONENTS
        for name, field in variables:
            visf.grid.add_pointdata(DataArray(name, scale_factor*field,
                vector_format=VF_LIST_OF_COMPONENTS))

        self.register_pathname(time, visf.get_head_pathname())





# silo ------------------------------------------------------------------------
class SiloMeshData(object):
    def __init__(self, dim, coords, element_groups):
        self.coords = coords

        from pylo import IntVector
        self.ndims = dim
        self.nodelist = IntVector()
        self.shapesizes = IntVector()
        self.shapecounts = IntVector()
        self.nshapetypes = 0
        self.nzones = 0

        for nodelist_size_estimate, eg in element_groups:
            poly_count = 0
            poly_length = None
            self.nodelist.reserve(len(self.nodelist) + nodelist_size_estimate)
            for polygon in eg:
                prev_nodelist_len = len(self.nodelist)
                for i in polygon:
                    self.nodelist.append(i)
                poly_count += 1
                poly_length = len(self.nodelist) - prev_nodelist_len

            if poly_count:
                self.shapesizes.append(poly_length)
                self.shapecounts.append(poly_count)
                self.nshapetypes += 1
                self.nzones += poly_count

    def put_mesh(self, silo, zonelist_name, mesh_name, mesh_opts):
        # put zone list
        silo.put_zonelist(zonelist_name, self.nzones, self.ndims, self.nodelist,
                self.shapesizes, self.shapecounts)

        silo.put_ucdmesh(mesh_name, self.ndims, [], self.coords, self.nzones,
                zonelist_name, None, mesh_opts)




class SiloVisualizer(Visualizer):
    def __init__(self, discr, pcontext=None):
        def generate_fine_elements(eg):
            ldis = eg.local_discretization
            smi = ldis.get_submesh_indices()
            for el, (el_start, el_stop) in zip(eg.members, eg.ranges):
                for element in smi:
                    yield [el_start+j for j in element]

        def generate_fine_element_groups():
            for eg in discr.element_groups:
                ldis = eg.local_discretization
                smi = ldis.get_submesh_indices()
                nodelist_size_estimate = len(eg.members) * len(smi) * len(smi[0])
                yield nodelist_size_estimate, generate_fine_elements(eg)

        def generate_coarse_elements(eg):
            for el in eg.members:
                yield el.vertex_indices

        def generate_coarse_element_groups():
            for eg in discr.element_groups:
                if eg.members:
                    nodelist_size_estimate = len(eg.members) \
                            * len(eg.members[0].vertex_indices)
                else:
                    nodelist_size_estimate = 0

                yield nodelist_size_estimate, generate_coarse_elements(eg)

        self.dim = discr.dimensions
        if self.dim != 1:
            self.fine_mesh = SiloMeshData(self.dim, 
                    numpy.asarray(discr.nodes.T, order="C"),
                    generate_fine_element_groups())
            self.coarse_mesh = SiloMeshData(self.dim, 
                    numpy.asarray(discr.mesh.points.T, order="C"),
                    generate_coarse_element_groups())
        else:
            self.xvals = numpy.asarray(discr.nodes.T, order="C")
            
        self.pcontext = pcontext

    def close(self):
        pass

    def make_file(self, pathname):
        """This function returns either a pylo.SiloFile or a
        pylo.ParallelSiloFile, depending on the ParallelContext
        under which we are running

        An extension of .silo is automatically appended to `pathname'.
        """
        if self.pcontext is None or len(self.pcontext.ranks) == 1:
            from pylo import SiloFile
            return SiloFile(pathname+".silo")
        else:
            from pylo import ParallelSiloFile
            return ParallelSiloFile(
                    pathname, 
                    self.pcontext.rank, self.pcontext.ranks)

    def add_data(self, silo, variables=[], scalars=[], vectors=[], expressions=[],
            time=None, step=None, scale_factor=1):
        if scalars or vectors:
            import warnings
            warnings.warn("`scalars' and `vectors' arguments are deprecated",
                    DeprecationWarning)
            variables = scalars + vectors

        from pylo import DB_NODECENT, DBOPT_DTIME, DBOPT_CYCLE

        # put mesh coordinates
        mesh_opts = {}
        if time is not None:
            mesh_opts[DBOPT_DTIME] = float(time)
        if step is not None:
            mesh_opts[DBOPT_CYCLE] = int(step)

        if self.dim == 1:
            for name, field in variables:
                if isinstance(field, list) and len(field) > 1:
                    from warnings import warn
                    warn("Silo visualization does not support vectors in 1D, ignoring '%s'" % name)
                else:
                    if isinstance(field, list):
                        field = field[0]
                    silo.put_curve(name, self.xvals, scale_factor*field, mesh_opts)
        else:
            self.fine_mesh.put_mesh(silo, "finezonelist", "finemesh", mesh_opts)
            self.coarse_mesh.put_mesh(silo, "coarsezonelist", "mesh", mesh_opts)

            from hedge.tools import log_shape

            # put data
            for name, field in variables:
                ls = log_shape(field)
                if ls != () and ls[0] > 1:
                    assert len(ls) == 1
                    silo.put_ucdvar(name, "finemesh", 
                            ["%s_comp%d" % (name, i) 
                                for i in range(ls[0])],
                            scale_factor*field, DB_NODECENT)
                else:
                    if ls != ():
                        field = field[0]
                    silo.put_ucdvar1(name, "finemesh", scale_factor*field, DB_NODECENT)

        if expressions:
            silo.put_defvars("defvars", expressions)




# tools -----------------------------------------------------------------------
def get_rank_partition(pcon, discr):
    vec = discr.volume_zeros()
    vec[:] = pcon.rank
    return vec



"""This benchmark informs the decision whether to do matching of periodic
face pairs by lookup of tuples of vertices or each vertex separately.

At the time of this writing, and under Python 2.5, tuplewise lookup has
about a 1-to-4 advantage.
"""

def main():
    from random import choice
    from time import time

    N = 50
    base_set = range(N)
    
    # generate correspondence
    corr = dict((b, choice(base_set)) for b in base_set)

    # generate tuples
    TS = 3
    N_TUP = 500000
    tups = [tuple(choice(base_set) for i in range(TS))
            for j in range(N_TUP)]

    tup_corr = {}
    for tup in tups:
        mapped = tuple(corr[t] for t in tup)
        tup_corr[tup] = mapped

    ITER = 500000
    start = time()
    for i in xrange(ITER):
        tup = choice(tups)
        mapped = tuple(corr[t] for t in tup)
    print "elwise", time()-start

    start = time()
    for i in xrange(ITER):
        tup = choice(tups)
        mapped = tup_corr[tup]
    print "tuplewise", time()-start

if __name__ == "__main__":
    main()



"""Hedge is the Hybrid'n'Easy Discontinuous Galerkin Environment."""

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""

# make sure numpy-to-C++ converters are available
import pyublas



# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




from __future__ import division
import numpy
import numpy.linalg as la




def main():
    from hedge.element import TetrahedralElement
    from hedge.timestep import RK4TimeStepper
    from hedge.mesh import make_ball_mesh, make_cylinder_mesh, make_box_mesh
    from hedge.visualization import \
            VtkVisualizer, \
            SiloVisualizer, \
            get_rank_partition
    from hedge.discretization import norm
    from hedge.tools import EOCRecorder, to_obj_array
    from math import sqrt, pi
    from analytic_solutions import \
            check_time_harmonic_solution, \
            RealPartAdapter, \
            SplitComplexAdapter, \
            CartesianAdapter, \
            CylindricalCavityMode, \
            RectangularWaveguideMode, \
            RectangularCavityMode
    from hedge.operators import MaxwellOperator
    from hedge.parallel import guess_parallelization_context

    pcon = guess_parallelization_context()

    epsilon0 = 8.8541878176e-12 # C**2 / (N m**2)
    mu0 = 4*pi*1e-7 # N/A**2.
    epsilon = 1*epsilon0
    mu = 1*mu0

    eoc_rec = EOCRecorder()

    cylindrical = False
    periodic = False

    if cylindrical:
        R = 1
        d = 2
        mode = CylindricalCavityMode(m=1, n=1, p=1,
                radius=R, height=d, 
                epsilon=epsilon, mu=mu)
        r_sol = CartesianAdapter(RealPartAdapter(mode))
        c_sol = SplitComplexAdapter(CartesianAdapter(mode))

        if pcon.is_head_rank:
            mesh = make_cylinder_mesh(radius=R, height=d, max_volume=0.01)
    else:
        if periodic:
            mode = RectangularWaveguideMode(epsilon, mu, (3,2,1))
            periodicity = (False, False, True)
        else:
            periodicity = None
        mode = RectangularCavityMode(epsilon, mu, (1,2,2))
        r_sol = RealPartAdapter(mode)
        c_sol = SplitComplexAdapter(mode)

        if pcon.is_head_rank:
            mesh = make_box_mesh(max_volume=0.01, periodicity=periodicity)

    if pcon.is_head_rank:
        mesh_data = pcon.distribute_mesh(mesh)
    else:
        mesh_data = pcon.receive_mesh()

    #for order in [1,2,3,4,5,6]:
    for order in [2,3,4,5,6]:
        discr = pcon.make_discretization(mesh_data, TetrahedralElement(order))

        vis = VtkVisualizer(discr, pcon, "em-%d" % order)

        mode.set_time(0)
        fields = to_obj_array(discr.interpolate_volume_function(r_sol))
        op = MaxwellOperator(discr, epsilon, mu, upwind_alpha=1,
                direct_flux=True)

        dt = discr.dt_factor(op.max_eigenvalue())
        final_time = 1e-9
        nsteps = int(final_time/dt)+1
        dt = final_time/nsteps

        if pcon.is_head_rank:
            print "---------------------------------------------"
            print "order %d" % order
            print "---------------------------------------------"
            print "dt", dt
            print "nsteps", nsteps
            print "#elements=", len(mesh.elements)

        #check_time_harmonic_solution(discr, mode, c_sol)
        #continue

        #from pylinear.toybox import write_gnuplot_sparsity_pattern
        #write_gnuplot_sparsity_pattern(
                #"fluxmat-%d.dat" % pcon.rank, op.n_jump[0].serial_flux_op.int_matrix)

        #return

        stepper = RK4TimeStepper()

        # diagnostics setup ---------------------------------------------------
        from pytools.log import LogManager, add_general_quantities, \
                add_simulation_quantities, add_run_info

        logmgr = LogManager("maxwell-%d.dat" % order, "w", pcon.communicator)
        add_run_info(logmgr)
        add_general_quantities(logmgr)
        add_simulation_quantities(logmgr, dt)
        discr.add_instrumentation(logmgr)
        stepper.add_instrumentation(logmgr)

        from pytools.log import IntervalTimer
        vis_timer = IntervalTimer("t_vis", "Time spent visualizing")
        logmgr.add_quantity(vis_timer)

        from hedge.log import EMFieldGetter, add_em_quantities
        field_getter = EMFieldGetter(op, locals(), "fields")
        add_em_quantities(logmgr, op, field_getter)
        
        logmgr.add_watches(["step.max", "t_sim.max", "W_field", "t_step.max"])

        # timestep loop -------------------------------------------------------
        t = 0
        for step in range(nsteps):
            logmgr.tick()

            if True:
                vis_timer.start()
                e, h = op.split_eh(fields)
                visf = vis.make_file("em-%d-%04d" % (order, step))
                vis.add_data(visf,
                        [ ("e", e), ("h", h), ],
                        time=t, step=step
                        )
                visf.close()
                vis_timer.stop()

            fields = stepper(fields, t, dt, op.rhs)
            t += dt

        logmgr.tick()
        logmgr.save()

        numpy.seterr('raise')
        mode.set_time(t)
        true_fields = to_obj_array(discr.interpolate_volume_function(r_sol))

        eoc_rec.add_data_point(order, norm(discr, fields-true_fields))

        print
        print eoc_rec.pretty_print("P.Deg.", "L2 Error")

if __name__ == "__main__":
    import cProfile as profile
    #profile.run("main()", "wave2d.prof")
    main()
